from datetime import datetime
from pathlib import Path
from typing import Optional, Dict, List

from phidata.asset.data_asset import DataAsset
from phidata.task.task import Task
from phidata.checks.check import Check
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.workspace import WorkspaceConfig
from phidata.types.context import RunContext
from pydantic import ValidationError

from phiterm.utils.cli_console import (
    print_subheading,
    print_heading,
    print_info,
    print_warning,
)
from phiterm.utils.log import logger
from phiterm.workflow.wf_enums import WorkflowEnv
from phiterm.workflow.wf_utils import (
    parse_workflow_file,
    parse_workflow_description,
)
from phiterm.workspace.phi_ws_data import PhiWsData


def run_workflow(
    wf_description: str,
    ws_data: PhiWsData,
    target_env: WorkflowEnv,
    target_dttm: datetime,
    dry_run: bool,
    detach: bool,
    target_app: Optional[str] = None,
    run_env_vars: Optional[Dict[str, str]] = None,
    run_params: Optional[Dict[str, str]] = None,
) -> None:
    """Runs the Phidata Workflow in the target environment.
    The run_workflow() function is called from `phi wf run`

    Step 1: Get the ws_root_path and workflows_dir
    Step 2: Get the workflow_file and workflow_name
    Step 3: Get the absolute workflow_file_path
    Step 4: Get DataProducts/Workflows/Tasks in the workflow_file_path
    Step 5: Run DataProducts/Workflows/Tasks in target runtime
    """

    if ws_data is None or ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return
    ws_config: WorkspaceConfig = ws_data.ws_config

    # Step 1: Get the ws_root_path and workflows_dir
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return
    ws_root_path: Path = ws_data.ws_root_path

    if ws_config.workflows_dir is None:
        logger.error("Workflow directory invalid")
        return
    workflows_dir: str = ws_config.workflows_dir

    # Step 2: Get the workflow_file and workflow_name
    # The workflow_file is the path of the file relative to the workflows_dir.
    workflow_file, workflow_name, task_name = parse_workflow_description(
        wf_description=wf_description,
        ws_root_path=ws_root_path,
        workflows_dir=workflows_dir,
    )
    # Validate workflow_file
    if workflow_file is None:
        logger.error("Workflow file not found")
        return

    print_info("File: {}".format(workflow_file))
    if workflow_name:
        print_info("Workflow: {}".format(workflow_name))
    if task_name:
        print_info("Task: {}".format(task_name))

    # Step 3: Get the absolute workflow_file_path
    workflow_file_path: Path
    try:
        workflow_file_path = (
            ws_root_path.joinpath(workflows_dir)
            .joinpath(workflow_file)
            .resolve(strict=True)
        )
    except FileNotFoundError as e:
        logger.error("Workflow file not found: {}".format(workflow_file))
        return

    # Step 4: Get DataProducts/Workflows/Tasks in the workflow_file_path
    dp_dict: Dict[str, DataProduct]
    wf_dict: Dict[str, Workflow]
    task_dict: Dict[str, Task]
    da_dict: Dict[str, DataAsset]
    dq_dict: Dict[str, Check]
    try:
        dp_dict, wf_dict, task_dict, da_dict, dq_dict = parse_workflow_file(
            workflow_file_path
        )
    except ValidationError as e:
        from phidata.utils.cli_console import print_validation_errors

        print_validation_errors(e.errors())
        print_info("Please fix and try again")
        return
    logger.debug(f"DataProducts: {dp_dict}")
    logger.debug(f"Workflows: {wf_dict}")
    logger.debug(f"Tasks: {task_dict}")
    logger.debug(f"DataAssets: {da_dict}")
    logger.debug(f"Checks: {dq_dict}")

    # Final run status
    run_status: bool = False
    # Build the RunContext
    run_context = RunContext(
        run_date=target_dttm,
        dry_run=dry_run,
        detach=detach,
        run_env=target_env.value,
        run_env_vars=run_env_vars,
        run_params=run_params,
    )

    # Step 5: Run DataProducts/Workflows/Tasks in target runtime

    # Run Workflow locally
    if target_env == WorkflowEnv.local:
        from phiterm.local.local_operator import (
            run_workflows_local,
            run_data_products_local,
        )

        # ** CASE 1: If a workflow_name or a task_name is provided
        if (workflow_name is not None and len(wf_dict) > 0) or (
            task_name is not None and len(task_dict) > 0
        ):
            run_status = run_workflows_local(
                workflow_file=workflow_file,
                run_context=run_context,
                ws_data=ws_data,
                workflows=wf_dict,
                workflow_name=workflow_name,
                tasks=task_dict,
                task_name=task_name,
            )
        # ** CASE 2: If a workflow_name or a task name is not provided
        #   and workflow_file contains DataProducts.
        #   Run the DataProducts
        elif dp_dict is not None and len(dp_dict) > 0:
            run_status = run_data_products_local(
                workflow_file=workflow_file,
                data_products=dp_dict,
                run_context=run_context,
                ws_data=ws_data,
            )
        # ** CASE 3: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   but workflow_file contains Workflows
        #   Run all Workflows
        elif wf_dict is not None and len(wf_dict) > 0:
            run_status = run_workflows_local(
                workflow_file=workflow_file,
                run_context=run_context,
                ws_data=ws_data,
                workflows=wf_dict,
            )
        # ** CASE 4: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   and workflow_file does NOT contain Workflows
        #   but workflow_file contains Tasks
        #   Run all Tasks
        elif task_dict is not None and len(task_dict) > 0:
            run_status = run_workflows_local(
                workflow_file=workflow_file,
                run_context=run_context,
                ws_data=ws_data,
                workflows=wf_dict,
                tasks=task_dict,
            )

    # Run Workflow in docker
    elif target_env == WorkflowEnv.dev:
        from phidata.docker.config import DockerConfig
        from phiterm.docker.docker_operator import (
            run_workflows_docker,
            run_data_products_docker,
        )
        from phidata.utils.prep_infra_config import prep_infra_config

        # Find the DockerConfig to use
        docker_configs: Optional[List[DockerConfig]] = ws_config.docker
        filtered_docker_config: Optional[DockerConfig] = None
        if docker_configs is not None and isinstance(docker_configs, list):
            if len(docker_configs) == 1:
                filtered_docker_config = docker_configs[0]
            else:
                for dc in docker_configs:
                    if dc.env == target_env.value:
                        filtered_docker_config = dc
                        break
        if filtered_docker_config is None:
            logger.error(f"No DockerConfig found for env: {target_env.value}")
            return

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update sub-configs data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down docker_env which is used to set the env variables
        #       when running the docker command
        ######################################################################

        docker_config_to_use: Optional[DockerConfig] = None
        _config = prep_infra_config(
            infra_config=filtered_docker_config,
            ws_config=ws_config,
        )
        if isinstance(_config, DockerConfig):
            docker_config_to_use = _config

        if docker_config_to_use is None:
            logger.error(f"No DockerConfig found for env: {target_env.value}")
            return

        # ** CASE 1: If a workflow_name or a task_name is provided
        if (workflow_name is not None and len(wf_dict) > 0) or (
            task_name is not None and len(task_dict) > 0
        ):
            # Because we are running this workflow/task as an airflow task,
            # we need to identify the DAG it belongs to.
            # Workflow DAGs are usually provided by their DataProducts
            # but can also have their own DAGs.
            dag_id_to_use = None

            # If a workflow is provided, lets find out if this workflow
            # is a part of a DataProduct. If it is, update the dag_id_to_use
            if workflow_name != "" and len(dp_dict) > 0:
                logger.debug(f"Finding DAG for workflow: {workflow_name}")
                if workflow_name not in wf_dict:
                    logger.error(
                        "Could not find '{}' in {}".format(
                            workflow_name, "[{}]".format(", ".join(wf_dict.keys()))
                        )
                    )
                    return

                # find the workflow to run
                workflow_to_run = wf_dict[workflow_name]
                logger.debug(f"Found workflow: {workflow_to_run}")

                # Check if the workflow is part of a DP
                # if it is, get the dag_id_to_use
                for dp_name, dp_obj in dp_dict.items():
                    dp_workflows = dp_obj.workflows
                    logger.debug(f"dp_workflows: {dp_workflows}")
                    if dp_workflows is not None and workflow_to_run in dp_workflows:
                        logger.debug(f"Using data product: {dp_name}")
                        dag_id_to_use = dp_obj.dag_id
                        logger.info(f"DAG.id: {dag_id_to_use}")

            # TODO: update this when testing `phi wf run dp::task`
            # # If a workflow is not provided, but a task is
            # # find the workflow for the task and use the dag_id from the workflow
            # elif task_name != "" and len(wf_dict) > 0:
            #     workflow = None
            #     if len(wf_dict) == 1:
            #         workflow = wf_dict.get(list(wf_dict.keys())[0])
            #     else:
            #         for wf_name, wf_obj in wf_dict.items():
            #             wf_tasks = {t.task_id: t for t in wf_obj.tasks}
            #             logger.debug(f"wf_tasks: {wf_tasks}")
            #             if wf_tasks is not None and task_name in wf_tasks:
            #                 workflow = wf_obj
            #
            #     if workflow is not None:
            #         logger.debug(f"Using workflow: {workflow.name}")
            #         dag_id_to_use = workflow.dag_id
            #         logger.debug(f"Using DAG: {dag_id_to_use}")

            run_status = run_workflows_docker(
                workflow_file=workflow_file,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
                workflows=wf_dict,
                workflow_name=workflow_name,
                task_name=task_name,
                dag_id=dag_id_to_use,
            )
        # ** CASE 2: If a workflow_name or a task name is not provided
        #   and workflow_file contains DataProducts.
        #   Run the DataProducts
        elif dp_dict is not None and len(dp_dict) > 0:
            run_status = run_data_products_docker(
                workflow_file=workflow_file,
                data_products=dp_dict,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
            )
        # ** CASE 3: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   but workflow_file contains Workflows
        #   Run the Workflows
        elif wf_dict is not None and len(wf_dict) > 0:
            run_status = run_workflows_docker(
                workflow_file=workflow_file,
                run_context=run_context,
                docker_config=docker_config_to_use,
                target_app=target_app,
                workflows=wf_dict,
            )
    # Run Workflow in staging k8s
    elif target_env == WorkflowEnv.stg:
        logger.error(f"WorkflowEnv: {target_env} not yet supported")
    # Run Workflow in prod k8s
    elif target_env == WorkflowEnv.prd:
        print_warning(
            f"Running workflows in {target_env.value} is in alpha, please verify results"
        )
        from phidata.k8s.config import K8sConfig
        from phiterm.k8s.k8s_operator import (
            run_workflows_k8s,
            run_data_products_k8s,
        )
        from phidata.utils.prep_infra_config import prep_infra_config

        k8s_configs: Optional[List[K8sConfig]] = ws_config.k8s
        filtered_k8s_config: Optional[K8sConfig] = None
        if k8s_configs is not None and isinstance(k8s_configs, list):
            if len(k8s_configs) == 1:
                filtered_k8s_config = k8s_configs[0]
            else:
                for kc in k8s_configs:
                    if kc.env == target_env.value:
                        filtered_k8s_config = kc
                        break
        if filtered_k8s_config is None:
            logger.error(f"No K8sConfig found for env: {target_env.value}")
            return

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update sub-configs data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down k8s_env which is used to set the env variables
        #       when running the k8s command
        ######################################################################

        k8s_config_to_use: Optional[K8sConfig] = None
        _config = prep_infra_config(
            infra_config=filtered_k8s_config,
            ws_config=ws_config,
        )
        if isinstance(_config, K8sConfig):
            k8s_config_to_use = _config

        if k8s_config_to_use is None:
            logger.error(f"No K8sConfig found for env: {target_env.value}")
            return

        # ** CASE 1: If a workflow_name is provided
        #   Run only the Workflow
        if workflow_name is not None and len(wf_dict) > 0:
            # Because this is a single workflow, we need to identify the DAG
            # it belongs to. Workflow DAGs are usually provided by their DataProducts.
            # But Workflows can independently have their own DAGs as well.
            # To run a single workflow, lets find out if this workflow
            #   is a part of a DataProduct
            #   or an independent workflow
            wf_dag_id = None
            logger.debug(f"Finding DAG for {workflow_name}")
            if len(dp_dict) > 0:
                if workflow_name not in wf_dict:
                    logger.error(
                        "Could not find '{}' in {}".format(
                            workflow_name, "[{}]".format(", ".join(wf_dict.keys()))
                        )
                    )
                else:
                    _workflow = wf_dict[workflow_name]
                    # logger.debug(f"Found workflow: {_workflow}")
                    for dp_name, dp_obj in dp_dict.items():
                        dp_workflows = dp_obj.workflows
                        # logger.debug(f"dp_workflows: {dp_workflows}")
                        if dp_workflows is not None and _workflow in dp_workflows:
                            # logger.debug(f"Found data product: {dp_name}")
                            wf_dag_id = dp_obj.dag_id
                            print_info(f"Workflow DAG: {wf_dag_id}")

            run_status = run_workflows_k8s(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                k8s_config=k8s_config_to_use,
                target_app=target_app,
                workflow_name=workflow_name,
                use_dag_id=wf_dag_id,
            )
        # ** CASE 2: If a workflow_name is not provided
        #   but workflow_file contains DataProducts.
        #   Run the DataProducts
        elif dp_dict is not None and len(dp_dict) > 0:
            run_status = run_data_products_k8s(
                workflow_file=workflow_file,
                data_products=dp_dict,
                run_context=run_context,
                k8s_config=k8s_config_to_use,
                target_app=target_app,
            )
        # ** CASE 3: If a workflow_name is not provided
        #   and workflow_file does NOT contain DataProducts
        #   but workflow_file contains Workflows
        #   Run the Workflows
        elif wf_dict is not None and len(wf_dict) > 0:
            run_status = run_workflows_k8s(
                workflow_file=workflow_file,
                workflows=wf_dict,
                run_context=run_context,
                k8s_config=k8s_config_to_use,
                target_app=target_app,
            )
    else:
        logger.error(f"WorkflowEnv: {target_env} not supported")

    # Report the run_status
    print_info("")
    if run_status:
        logger.info("Success")
    else:
        logger.error("Fail")


def print_workflow_details(wf: Workflow) -> None:
    # Print workflow objects which are not imported in the file
    # print workflow tasks
    print_subheading(f"\nWorkflow: {wf.name}")
    family_tree: Dict[str, Dict[str, List[str]]] = wf.family_tree_dict
    if wf.tasks is not None:
        print_subheading(f"  Tasks: {len(wf.tasks)}")
        for task in wf.tasks:
            print_info("   * {}".format(task.name))
            if task.task_id in family_tree:
                task_tree = family_tree[task.task_id]
                for relation_type, relatives in task_tree.items():
                    logger.debug("{}: {}".format(relation_type, relatives))

    # print workflow pre-checks
    if wf.pre_checks is not None:
        print_subheading(f"  Pre-Checks: {len(wf.pre_checks)}")
        for pre_checks in wf.pre_checks:
            print_info("   * {}".format(pre_checks.name))
    # print workflow post-checks
    if wf.post_checks is not None:
        print_subheading(f"  Post-Checks: {len(wf.post_checks)}")
        for post_check in wf.post_checks:
            print_info("   * {}".format(post_check.name))
    # print workflow inputs
    if wf.inputs is not None:
        print_subheading(f"  Inputs: {len(wf.inputs)}")
        for _input in wf.inputs:
            print_info("   * {}".format(_input.name))
    # print workflow outputs
    if wf.outputs is not None:
        print_subheading(f"  Outputs: {len(wf.outputs)}")
        for output in wf.outputs:
            print_info("   * {} : {}".format(output.name, output.__class__.__name__))
    # if wf.tree is not None:
    #     print_subheading(f"  Tree:")
    #     print_info(json.dumps(wf.tree, indent=2))


def list_workflows(
    wf_description: str,
    ws_data: PhiWsData,
) -> None:
    if ws_data is None or ws_data.ws_config is None:
        logger.error("WorkspaceConfig invalid")
        return
    ws_config: WorkspaceConfig = ws_data.ws_config

    # Step 1: Get the ws_root_path and workflows_dir
    if ws_data.ws_root_path is None:
        logger.error("Workspace directory invalid")
        return
    ws_root_path: Path = ws_data.ws_root_path

    if ws_config.workflows_dir is None:
        logger.error("Workflow directory invalid")
        return
    workflows_dir: str = ws_config.workflows_dir

    # Step 2: Get the workflow_file and workflow_name
    # The workflow_file is the path of the file relative to the workflows_dir.
    print_subheading(f"Parsing: {wf_description}")
    workflow_file, workflow_name, task_name = parse_workflow_description(
        wf_description=wf_description,
        ws_root_path=ws_root_path,
        workflows_dir=workflows_dir,
    )
    # Validate workflow_file
    if workflow_file is None:
        logger.error("Workflow file not found")
        return

    print_info("File: {}".format(workflow_file))
    if workflow_name:
        print_info("Workflow: {}".format(workflow_name))
    if task_name:
        print_info("Task: {}".format(task_name))

    # Step 3: Get the absolute workflow_file_path
    workflow_file_path: Path
    try:
        workflow_file_path = (
            ws_root_path.joinpath(workflows_dir)
            .joinpath(workflow_file)
            .resolve(strict=True)
        )
    except FileNotFoundError as e:
        logger.error("Workflow file not found: {}".format(workflow_file))
        return

    # Step 4: Get DataProducts/Workflows in the workflow_file_path
    dp_dict: Dict[str, DataProduct]
    wf_dict: Dict[str, Workflow]
    task_dict: Dict[str, Task]
    da_dict: Dict[str, DataAsset]
    dq_dict: Dict[str, Check]
    try:
        dp_dict, wf_dict, task_dict, da_dict, dq_dict = parse_workflow_file(
            workflow_file_path
        )
    except ValidationError as e:
        from phidata.utils.cli_console import print_validation_errors

        print_validation_errors(e.errors())
        print_info("Please fix and try again")
        return

    # Step 5: Print :)

    if dp_dict is not None and len(dp_dict) > 0:
        print_heading(f"\nDataProducts: {len(dp_dict)}")
        for key, dp in dp_dict.items():
            print_info(" * {}".format(dp.name))

    if wf_dict is not None and len(wf_dict) > 0:
        print_heading(f"\nWorkflows: {len(wf_dict)}")
        if workflow_name is None:
            for key, wf in wf_dict.items():
                print_workflow_details(wf)
        else:
            wf = wf_dict.get(workflow_name, None)
            if wf is None:
                logger.error(
                    "Could not find '{}' in {}".format(
                        workflow_name, "[{}]".format(", ".join(wf_dict.keys()))
                    )
                )
            else:
                print_workflow_details(wf)

    if task_dict is not None and len(task_dict) > 0:
        print_heading(f"\nTasks: {len(task_dict)}")
        for key, task in task_dict.items():
            print_info(" * {}".format(task.name))

    if dq_dict is not None and len(dq_dict) > 0:
        print_heading(f"\nDQ Checks: {len(dq_dict)}")
        for key, dq in dq_dict.items():
            print_info(" * {}".format(dq.name))

    if da_dict is not None and len(da_dict) > 0:
        print_heading(f"\nDataAssets: {len(da_dict)}")
        for key, da in da_dict.items():
            print_info("{}: {}".format(da.__class__.__name__, da.name))



from typing import List, Optional

import httpx

from phiterm.backend_api.api_client import BackendApi, get_authenticated_client
from phiterm.backend_api.api_utils import response_is_invalid
from phiterm.utils.cli_console import log_network_error_msg
from phiterm.utils.log import logger
from phiterm.workspace.ws_schemas import (
    UpdateWorkspace,
    UpsertWorkspaceFromCli,
    WorkspaceSchema,
)


def convert_ws_response_to_schema(
    resp: httpx.Response,
) -> Optional[WorkspaceSchema]:
    ws_dict = resp.json()
    if ws_dict is None or not isinstance(ws_dict, dict):
        logger.debug("Could not parse ws data: {}".format(ws_dict))
        return None

    if resp.status_code == httpx.codes.OK:
        return WorkspaceSchema.from_dict(ws_dict)

    return None


def get_primary_workspace() -> Optional[WorkspaceSchema]:
    logger.debug("--o-o-- Get primary workspace")
    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            r: httpx.Response = api.get(BackendApi.workspace_read_primary)
            if response_is_invalid(r):
                return None
        except httpx.NetworkError:
            log_network_error_msg()
            return None

        # logger.debug("url: {}".format(r.url))
        # logger.debug("status: {}".format(r.status_code))
        # logger.debug("json: {}".format(r.json()))
        return convert_ws_response_to_schema(r)


def get_available_workspaces() -> Optional[List[WorkspaceSchema]]:
    logger.debug("--o-o-- Get available workspaces")
    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            r: httpx.Response = api.get(BackendApi.workspaces_read_available)
            if response_is_invalid(r):
                return None
        except httpx.NetworkError:
            log_network_error_msg()
            return None

        # logger.debug("url: {}".format(r.url))
        # logger.debug("status: {}".format(r.status_code))
        # logger.debug("json: {}".format(r.json()))

        ws_list_dict = r.json()
        # logger.debug("ws_list_dict type: {}".format(type(ws_list_dict)))
        if ws_list_dict is None or not isinstance(ws_list_dict, list):
            logger.debug("no workspaces available")
            return []
        # convert ws_list_dict to List[WorkspaceSchema] and return
        if r.status_code == httpx.codes.OK:
            ws_list: List[WorkspaceSchema] = []
            for ws_dict in ws_list_dict:
                if not isinstance(ws_dict, dict):
                    logger.debug("Could not parse {}".format(ws_dict))
                    continue
                ws_list.append(WorkspaceSchema.from_dict(ws_dict))

            return ws_list

    return None


def upsert_workspace(
    ws_upsert: UpsertWorkspaceFromCli,
) -> Optional[WorkspaceSchema]:
    logger.debug("--o-o-- Upsert workspace")
    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        logger.debug("User not authenticated")
        return None
    with authenticated_client as api:
        try:
            # logger.debug("data: {}".format(ws_upsert.dict()))
            r: httpx.Response = api.post(
                BackendApi.workspace_upsert_from_cli, json=ws_upsert.dict()
            )
            if response_is_invalid(r):
                return None
        except httpx.NetworkError:
            log_network_error_msg()
            return None

        # logger.debug("url: {}".format(r.url))
        # logger.debug("status: {}".format(r.status_code))
        # logger.debug("json: {}".format(r.json()))
        return convert_ws_response_to_schema(r)


def update_workspace(
    ws_update: UpdateWorkspace,
) -> Optional[WorkspaceSchema]:
    logger.debug("--o-o-- Update workspace")
    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None
    with authenticated_client as api:
        try:
            r: httpx.Response = api.post(
                BackendApi.workspace_update, json=ws_update.dict()
            )
            if response_is_invalid(r):
                return None
        except httpx.NetworkError:
            log_network_error_msg()
            return None

        logger.debug("url: {}".format(r.url))
        logger.debug("status: {}".format(r.status_code))
        logger.debug("json: {}".format(r.json()))
        return convert_ws_response_to_schema(r)



from typing import Optional

from phiterm.conf.constants import PHI_CREDS_PATH
from phiterm.utils.common import pickle_object_to_file, unpickle_object_from_file
from phiterm.utils.log import logger


class PhiCliCreds:
    def __init__(self, auth_token: str):
        self._auth_token = auth_token

    @property
    def auth_token(self) -> str:
        return self._auth_token


def save_auth_token(auth_token: str):
    # logger.debug(f"Storing {auth_token} to {str(PHI_CREDS_PATH)}")
    creds = PhiCliCreds(auth_token)
    pickle_object_to_file(creds, PHI_CREDS_PATH)


def read_auth_token() -> Optional[str]:
    # logger.debug(f"Reading token from {str(PHI_CREDS_PATH)}")
    creds: PhiCliCreds = unpickle_object_from_file(PHI_CREDS_PATH)
    return creds.auth_token



# """Phi GCP Cli
#
# This is the entrypoint for the `phi gcp` commands
# """
#
# from typing import Optional
#
# import typer
#
# from phi import schemas
# from phiterm.gcp import gcp_operator
# from phiterm.conf.phi_conf import PhiConf
# from phiterm.utils.common import conf_not_available_msg, primary_ws_not_available_msg
#
# _phi_gcp_help_text = """
# \b
# Commands to manage the GCP Project for your primary workspaces.
#
# \b
# Main Commands:
# $ phi gcp auth   -> Authenticates the GCP Project used by your primary ws
# $ phi gcp status -> Prints the GCP Project status for the primary ws
# """
# app = typer.Typer(name="gcp", help=_phi_gcp_help_text)
#
#
# @app.command(short_help="Authenticate the GCP Project used by your primary ws")
# def auth():
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     gcp_operator.authenticate_gcp_project_for_ws(primary_ws, config)
#
#
# if __name__ == "__main__":
#     app()



from httpx import Client, AsyncClient
from typing import Optional

from phiterm.conf.constants import (
    APP_NAME,
    APP_VERSION,
    BACKEND_API_URL,
    PHI_AUTH_TOKEN_COOKIE,
    PHI_AUTH_TOKEN_HEADER,
)
from phiterm.utils.log import logger

base_headers = {
    "user-agent": f"{APP_NAME}/{APP_VERSION}",
    "Content-Type": "application/json",
}


def get_authenticated_client() -> Optional[Client]:
    """Returns an instance of httpx.Client which with preconfigured auth and base url"""

    from phiterm.conf.auth import read_auth_token

    try:
        auth_token: Optional[str] = read_auth_token()
    except Exception as e:
        # logger.warning(f"Failed to read auth token: {e}")
        return None

    if auth_token is None:
        return None

    authenticated_headers = base_headers.copy()
    authenticated_headers[PHI_AUTH_TOKEN_HEADER] = auth_token
    return Client(
        base_url=BACKEND_API_URL,
        headers=authenticated_headers,
        cookies={PHI_AUTH_TOKEN_COOKIE: auth_token},
        timeout=60,
    )


def get_async_client() -> Optional[AsyncClient]:
    """Returns an instance of httpx.AsyncClient which with preconfigured auth and base url"""

    from phiterm.conf.auth import read_auth_token

    try:
        auth_token: Optional[str] = read_auth_token()
    except Exception:
        # logger.warning(f"Failed to read auth token: {e}")
        return None

    if auth_token is None:
        return None

    authenticated_headers = base_headers.copy()
    authenticated_headers[PHI_AUTH_TOKEN_HEADER] = auth_token
    return AsyncClient(
        base_url=BACKEND_API_URL,
        headers=authenticated_headers,
        cookies={PHI_AUTH_TOKEN_COOKIE: auth_token},
        timeout=60,
    )



from typing import List, Optional

import json
from httpx import Client, Response, NetworkError, codes

from phiterm.api.routes import ApiRoutes
from phiterm.api.handler import invalid_response
from phiterm.conf.constants import BACKEND_API_URL
from phiterm.schemas.user import UserSchema
from phiterm.schemas.workspace import WorkspaceSchema, WorkspaceActionData
from phiterm.utils.cli_console import log_network_error_msg
from phiterm.utils.log import logger


def parse_ws_response(
    resp: Response,
) -> Optional[WorkspaceSchema]:
    ws_dict = resp.json()
    if ws_dict is None or not isinstance(ws_dict, dict):
        logger.debug(f"Could not parse ws: {ws_dict}")
        return None

    # logger.debug("url: {}".format(resp.url))
    # logger.debug("status: {}".format(resp.status_code))
    # logger.debug("response: {}".format(resp.json()))

    if resp.status_code == codes.OK:
        return WorkspaceSchema.parse_obj(ws_dict)

    return None


def workspace_ping() -> bool:
    from phiterm.api.client import base_headers

    logger.debug("--o-o-- Pinging workspace api")

    with Client(base_url=BACKEND_API_URL, headers=base_headers, timeout=60) as api:
        try:
            r: Response = api.get(ApiRoutes.WORKSPACE_PING)
            if invalid_response(r):
                return False
        except NetworkError as e:
            log_network_error_msg()
            return False

        logger.debug("status: {}".format(r.status_code))
        logger.debug("headers: {}".format(r.headers))
        logger.debug("cookies: {}".format(r.cookies))
        logger.debug("url: {}".format(r.url))
        logger.debug("json: {}".format(r.json()))
        if r.status_code == codes.OK:
            return True

    return False


def get_primary_workspace(user: UserSchema) -> Optional[WorkspaceSchema]:
    from phiterm.api.client import get_authenticated_client

    logger.debug("--o-o-- Get primary workspace")

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            r: Response = api.post(
                ApiRoutes.WORKSPACE_READ_PRIMARY, data=user.json(exclude_none=True)  # type: ignore
            )
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        return parse_ws_response(r)


def get_available_workspaces(user: UserSchema) -> Optional[List[WorkspaceSchema]]:
    from phiterm.api.client import get_authenticated_client

    logger.debug("--o-o-- Get available workspaces")

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            r: Response = api.post(
                ApiRoutes.WORKSPACES_READ_AVAILABLE, data=user.json(exclude_none=True)  # type: ignore
            )
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        ws_list_dict = r.json()
        if ws_list_dict is None or not isinstance(ws_list_dict, list):
            logger.debug("No workspaces received")
            return []

        # convert ws_list_dict to List[WorkspaceSchema] and return
        if r.status_code == codes.OK:
            ws_list: List[WorkspaceSchema] = []
            for ws_dict in ws_list_dict:
                if not isinstance(ws_dict, dict):
                    logger.debug("Could not parse {}".format(ws_dict))
                    continue
                ws_list.append(WorkspaceSchema.parse_obj(ws_dict))

            return ws_list

    return None


def create_workspace(
    user: UserSchema,
    workspace: WorkspaceSchema,
) -> Optional[WorkspaceSchema]:
    from phiterm.api.client import get_authenticated_client

    logger.debug("--o-o-- Create workspace")

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            body = {
                "workspace_data": workspace.dict(exclude_none=True),
                "user_data": user.dict(exclude_none=True),
            }
            body_json = json.dumps(body)
            # logger.debug(f"body_json: {body_json})")
            r: Response = api.post(ApiRoutes.WORKSPACE_CREATE, json=body)
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        return parse_ws_response(r)


def update_workspace(
    user: UserSchema,
    workspace: WorkspaceSchema,
) -> Optional[WorkspaceSchema]:
    from phiterm.api.client import get_authenticated_client

    logger.debug("--o-o-- Update workspace")

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            body = {
                "workspace_data": workspace.dict(exclude_none=True),
                "user_data": user.dict(exclude_none=True),
            }
            body_json = json.dumps(body)
            # logger.debug(f"body_json: {body_json})")
            r: Response = api.post(ApiRoutes.WORKSPACE_UPDATE, json=body)
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        return parse_ws_response(r)


def update_primary_workspace(
    user: UserSchema,
    workspace: WorkspaceSchema,
) -> Optional[WorkspaceSchema]:
    from phiterm.api.client import get_authenticated_client

    logger.debug(f"--o-o-- Update primary workspace to: {workspace.ws_name}")

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return None

    with authenticated_client as api:
        try:
            body = {
                "workspace_data": {
                    "update_primary_ws_id": workspace.id_workspace,
                    "update_primary_ws_name": workspace.ws_name,
                },
                "user_data": user.dict(exclude_none=True),
            }
            body_json = json.dumps(body)
            # logger.debug(f"body_json: {body_json})")
            r: Response = api.post(ApiRoutes.WORKSPACE_UPDATE_PRIMARY, json=body)
            if invalid_response(r):
                return None
        except NetworkError:
            log_network_error_msg()
            return None

        return parse_ws_response(r)


def delete_workspaces_api(
    user: Optional[UserSchema],
    workspaces_to_delete: List[str],
) -> bool:
    from phiterm.api.client import get_authenticated_client

    logger.debug("--o-o-- Deleting workspaces")

    if user is None:
        logger.error("User invalid")
        return False

    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        return False

    with authenticated_client as api:
        try:
            body = {
                "workspaces": workspaces_to_delete,
                "user_data": user.dict(exclude_none=True),
            }
            body_json = json.dumps(body)
            logger.debug(f"body_json: {body_json})")
            r: Response = api.post(ApiRoutes.WORKSPACE_DELETE, json=body)
            if invalid_response(r):
                return False
        except NetworkError:
            log_network_error_msg()
            return False

    return True



from typing import List, Optional

from pydantic import BaseModel


class ApiResponseData(BaseModel):
    status: str = "fail"
    message: str = "invalid request"
    message_log: Optional[List[str]] = None



# """Phi GKE Cli
#
# This is the entrypoint for the `phi gke` commands
# """
#
# from typing import Optional
#
# import typer
#
# from phi import schemas
# from phiterm.gcp.gke import gke_operator
# from phiterm.conf.phi_conf import PhiConf
# from phiterm.utils.common import (
#     conf_not_available_msg,
#     pprint_info,
#     pprint_heading,
#     pprint_info,
#     pprint_status,
#     primary_ws_not_available_msg,
# )
#
# _phi_gke_help_text = """
# \b
# Commands to manage GKE Clusters for your primary workspaces.
#
# \b
# Main Commands:
# $ phi gke create-cluster   -> Creates the GKE Cluster used by your primary ws
# $ phi gke delete-cluster   -> Deletes the GKE Cluster used by your primary ws
# $ phi gke status           -> Prints the GKE Cluster for your primary ws
# """
# app = typer.Typer(name="gke", help=_phi_gke_help_text)
#
#
# @app.command(short_help="Creates the GKE Cluster used by your primary ws")
# def create_cluster(
#     refresh: bool = typer.Option(
#         False,
#         "-r",
#         "--refresh",
#         help="Skips cache and refreshes the cluster",
#         show_default=True,
#     ),
# ):
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     gke_cluster: Optional[
#         schemas.GKEClusterSchema
#     ] = gke_operator.get_or_create_gke_cluster(primary_ws, config, refresh)
#     if gke_cluster:
#         gcloud_cmd = (
#             f"gcloud container clusters get-credentials {gke_cluster.cluster_name}"
#         )
#         pprint_info(f"To authenticate with gcloud, please run:\n\n\t{gcloud_cmd}\n")
#
#
# @app.command(short_help="Deletes the GKE Cluster for the primary workspace")
# def delete_cluster():
#
#     from phiterm.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     gke_cluster_deleted: bool = gke_operator.delete_gke_cluster_if_exists(
#         primary_ws, config
#     )
#     if gke_cluster_deleted:
#         pprint_info("GKE Cluster deleted")
#
#
# @app.command(short_help="Print the GKE status for the primary workspace")
# def status():
#
#     from phiterm.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     pprint_heading("GKE Cluster Status")
#     pprint_status("WorkspaceSchema: {}".format(primary_ws.name))
#     gke_operator.print_gke_cluster_status(primary_ws, config)
#
#
# if __name__ == "__main__":
#     app()



"""Phi Databox Cli

This is the entrypoint for the `phi dx` commands
"""

from typing import Optional

import typer

from phiterm.utils.cli_console import (
    print_conf_not_available_msg,
    print_active_workspace_not_available,
    print_available_workspaces,
)
from phiterm.utils.log import logger, set_log_level_to_debug
from phiterm.workspace.ws_enums import WorkspaceEnv

dx_app = typer.Typer(
    name="dx",
    short_help="Run commands in databox",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@dx_app.command(
    short_help="Run command in databox",
    options_metavar="\b",
    no_args_is_help=True,
)
def run(
    command: str = typer.Argument(
        ..., help="Command to run.", metavar="[command]", show_default=False
    ),
    env_filter: Optional[str] = typer.Option(
        "dev",
        "-e",
        "--env",
        metavar="",
        help="The environment to use",
    ),
    app_filter: Optional[str] = typer.Option(
        None,
        "-a",
        "--app",
        metavar="",
        help="The App to run the command in. Default: databox",
        show_default=False,
        hidden=True,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    from phiterm.conf.phi_conf import PhiConf, PhiWsData
    from phiterm.databox.databox_operator import run_command
    from phiterm.utils.load_env import load_env

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    # Load environment from .env
    load_env(dotenv_dir=active_ws_data.ws_root_path)

    target_env: Optional[str] = None
    target_app: Optional[str] = None

    if env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter
    if target_env is None:
        target_env = (
            active_ws_data.ws_config.default_env if active_ws_data.ws_config else "dev"
        )

    if app_filter is not None and isinstance(app_filter, str):
        target_app = app_filter

    logger.debug("Running command")
    logger.debug(f"\tcommand      : {command}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_app   : {target_app}")
    run_command(
        command=command,
        ws_data=active_ws_data,
        target_env=target_env,
        target_app=target_app,
    )



from phiterm.utils.enums import ExtendedEnum


class UserAuthProviderEnum(ExtendedEnum):
    email_pass = "email_pass"
    google = "google"
    github = "github"
    gitlab = "gitlab"
    linkedin = "linkedin"
    aws = "aws"
    api = "api"


class VersionControlProviderEnum(ExtendedEnum):
    GITHUB = "GITHUB"
    GITLAB = "GITLAB"
    BITBUCKET = "BITBUCKET"


class UserPermissions(ExtendedEnum):
    read_user = "read_user"
    edit_user = "edit_user"
    read_workspace = "read_workspace"
    edit_workspace = "edit_workspace"
    workspace_admin = "workspace_admin"



from pathlib import Path
from typing import Optional, Dict

from phiterm.utils.log import logger


def read_pyproject_phidata(pyproject_file: Path) -> Optional[Dict]:
    logger.debug(f"Reading {pyproject_file}")
    try:
        import tomli

        pyproject_dict = tomli.loads(pyproject_file.read_text())
        # logger.debug(f"pyproject_dict: {pyproject_dict}")
        phidata_conf = pyproject_dict.get("tool", {}).get("phidata", None)
        if phidata_conf is not None and isinstance(phidata_conf, dict):
            return phidata_conf
    except Exception as e:
        logger.error(f"Could not read {pyproject_file}")
        logger.error(e)
    return None



from pathlib import Path
from typing import List, Optional, cast, Dict

import git
import typer

from phidata.infra.config import InfraConfig
from phidata.workspace import WorkspaceConfig

from phiterm.conf.constants import DEFAULT_WS_NAME
from phiterm.conf.phi_conf import PhiConf, PhiWsData
from phiterm.schemas.workspace import WorkspaceSchema
from phiterm.utils.cli_console import (
    print_conf_not_available_msg,
    print_heading,
    print_info,
    print_subheading,
)
from phiterm.utils.log import logger
from phiterm.workspace.ws_enums import (
    WorkspaceConfigType,
    WorkspaceStarterTemplate,
)
from phiterm.workspace.exceptions import WorkspaceException

# List of reserved workspace names
RESERVED_WS_NAMES = ["root", "bedi", "ashpreet", "ashpreetbedi"]
TEMPLATE_TO_NAME_MAP: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.ml_app: "ml-app",
    WorkspaceStarterTemplate.ai_app: "ai-app",
    WorkspaceStarterTemplate.duck_gpt: "duck-gpt",
    WorkspaceStarterTemplate.aws_dp: "data-platform",
    WorkspaceStarterTemplate.aws_spark_dp: "spark-data-platform",
    WorkspaceStarterTemplate.aws_snowflake_dp: "snowflake-data-platform",
}
TEMPLATE_TO_REPO_MAP: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.ml_app: "https://github.com/phidatahq/ml-app.git",
    WorkspaceStarterTemplate.ai_app: "https://github.com/phidatahq/ai-app.git",
    WorkspaceStarterTemplate.duck_gpt: "https://github.com/phidatahq/duck-gpt.git",
    WorkspaceStarterTemplate.aws_dp: "https://github.com/phidatahq/aws-dp-template.git",
    WorkspaceStarterTemplate.aws_spark_dp: "https://github.com/phidatahq/aws-spark-dp-template.git",
    WorkspaceStarterTemplate.aws_snowflake_dp: "https://github.com/phidatahq/aws-snowflake-dp-template.git",
}


def create_new_workspace(ws_name: str, template: WorkspaceStarterTemplate) -> bool:
    """Creates a new workspace for the user.
    This function clones a phidata template on the users machine at the path:
        cwd/ws_name
    cwd: current working dir - where the command was run from.

    Steps:
    1. Validate ws_name and if we can create a folder with that name
    2. Clone the starter template in cwd/ws_name
    3. Delete the .git folder
    4. Provide the user with steps to add a remote origin for this workspace

    NOTE: till the user runs `phi ws setup` the workspace is not registered with phidata
        and can only be used locally
    """
    from shutil import copytree
    from phiterm.utils.filesystem import rmdir_recursive
    from phiterm.workspace.ws_utils import get_ws_config_dir_path
    from phiterm.utils.git import GitCloneProgress

    current_dir: Path = Path(".").resolve()

    # Phidata should be initialized before creating a workspace
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    # Error if the workspace name provided is a reserved name
    if ws_name in RESERVED_WS_NAMES:
        logger.error(f"The name `{ws_name}` is reserved")
        return False

    # Check if a workspace with the same name exists
    _existing_ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_name(ws_name)
    if _existing_ws_data is not None:
        logger.error(f"Found existing record for a workspace: {ws_name}")
        delete_existing_ws_data = typer.confirm(
            "Replace the existing record?", default=True
        )
        if delete_existing_ws_data:
            phi_conf.delete_ws(ws_name)
        else:
            return False

    # Check if we can create the workspace in the current dir
    ws_root_path: Path = current_dir.joinpath(ws_name)
    if ws_root_path.exists():
        logger.error(
            f"Directory {ws_root_path} exists, please delete files manually or choose another name for workspace"
        )
        return False

    print_info(f"Creating {str(ws_root_path)}")
    # Clone the starter repo for this cloud in the new directory
    repo_to_clone = TEMPLATE_TO_REPO_MAP.get(template)
    logger.debug("Cloning: {}".format(repo_to_clone))
    try:
        _cloned_git_repo: git.Repo = git.Repo.clone_from(
            repo_to_clone, str(ws_root_path), progress=GitCloneProgress()  # type: ignore
        )
    except Exception as e:
        logger.error(e)
        return False

    # Remove existing .git folder
    _dot_git_folder = ws_root_path.joinpath(".git")
    _dot_git_exists = _dot_git_folder.exists()
    if _dot_git_exists:
        logger.debug(f"Deleting {_dot_git_folder}")
        try:
            _dot_git_exists = not rmdir_recursive(_dot_git_folder)
        except Exception as e:
            logger.warning(f"Failed to delete {_dot_git_folder}: {e}")
            logger.info("Please delete the .git folder manually")
            pass

    phi_conf.add_new_ws_to_config(
        ws_name=ws_name,
        ws_root_path=ws_root_path,
    )

    try:
        ws_config_dir_path: Path = get_ws_config_dir_path(ws_root_path)
        ws_config_secrets_dir = ws_config_dir_path.joinpath("secrets").resolve()
        ws_config_example_secrets_dir = ws_config_dir_path.joinpath(
            "example_secrets"
        ).resolve()

        print_info(f"Creating {str(ws_config_secrets_dir)}")
        copytree(
            str(ws_config_example_secrets_dir),
            str(ws_config_secrets_dir),
        )
    except Exception as e:
        logger.warning(f"Could not create workspace/secrets: {e}")
        logger.warning(
            "Please manually copy workspace/example_secrets to workspace/secrets"
        )

    print_info(f"Your new workspace is available at {str(ws_root_path)}\n")
    return setup_workspace(ws_root_path=ws_root_path)


def create_workspace_using_input():
    """Takes input from the user and calls create_new_workspace()"""

    from phiterm.utils.common import is_empty, str_to_int

    # Display available starter templates and ask user to select one
    print_info("Input Template Number (default: ai-app)")
    templates = WorkspaceStarterTemplate.values_list()
    for idx, prvdr in enumerate(templates, start=1):
        print_info("  [{}] {}".format(idx, prvdr))

    # Get starter template from the user
    template_inp_raw = input(": ")
    # Convert input to int value.
    template_inp = str_to_int(template_inp_raw)
    template: WorkspaceStarterTemplate = WorkspaceStarterTemplate.ml_app

    if template_inp is not None:
        try:
            template = cast(
                WorkspaceStarterTemplate,
                WorkspaceStarterTemplate.from_str(templates[template_inp - 1]),
            )
        except Exception as e:
            logger.error(e)
    logger.debug("Selected: {}".format(template.value))

    # Get workspace name from user
    default_ws_name = TEMPLATE_TO_NAME_MAP.get(template, DEFAULT_WS_NAME)
    ws_name_inp_raw = input(f"Workspace name (default: {default_ws_name}): ")
    if is_empty(ws_name_inp_raw):
        ws_name_inp_raw = default_ws_name

    print_info(
        f"Creating workspace {ws_name_inp_raw} using the {template.value} template\n"
    )
    create_new_workspace(ws_name_inp_raw, template)


def initialize_workspace() -> None:
    """Initialize a phidata workspace.

    This function is called by `phi ws init` to create a new workspace.
    """
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    print_heading("New phidata workspace")
    create_workspace_using_input()


def setup_workspace(ws_root_path: Path) -> bool:
    """Setup a phidata workspace at directory: `ws_root_path`.
    This is the catchall function for a workspace. Run it in a directory and it
    should figure everything out.

    1. Validate pre-requisites
    1.1 Check ws_root_path is valid
    1.2 Check PhiConf is valid
    1.3 Validate PhiWsData is available
        Get the ws_data using the ws_root_path
        If ws_data is None, read the ws_config_file_path to map it to an available workspace
            This may happen if the user ran `phi init -r`
        If the user cloned this directory directly, there will be no record of this workspace in the PhiConf
        In that case, we manually add the ws to PhiConf and process the ws_config_file_path and setup this workspace.
    1.4 Check if this is the active workspace or if it needs to be set as active
    1.5 Check if remote origin is available

    2. Create or Update WorkspaceSchema
    If a ws_schema exists for this workspace, this workspace has been setup before
    2.1 Create WorkspaceSchema for a NEWLY CREATED WORKSPACE
        If ws_schema is None, this is a NEWLY CREATED WORKSPACE.
        i.e. A new workspace was created using create_new_workspace() <- which is called using `phi ws init`.
        This ws does not have a ws_schema, make an api call to create one.
        If the api call returns None, then we could not create the workspace.
    2.2 Update WorkspaceSchema if ws_schema exists
    2.3 Set as active if needed
    2.4 Refresh PhiWsData

    3. Complete Workspace setup
    3.1 Validate ws_config_file_path
    3.2 Validate PhiWsData
        Reaching here implies a all ws_data is available and updated.
        Validate the ws_data and print any pending setup actions.

    `phi ws setup` is a generic catch-all function. It should handle errors graciously
    and provide "how to fix" messages and "next steps" to get the user running.
    """
    from phiterm.utils.git import get_remote_origin_for_dir
    from phiterm.workspace.ws_utils import print_howtofix_pending_actions

    print_heading("Running workspace setup\n")

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check ws_root_path is valid
    ######################################################
    _ws_is_valid: bool = (
        ws_root_path is not None and ws_root_path.exists() and ws_root_path.is_dir()
    )
    if not _ws_is_valid:
        logger.error("Invalid directory: {}".format(ws_root_path))
        return False

    ######################################################
    # 1.2 Check PhiConf is valid
    ######################################################
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    ######################################################
    # 1.3 Validate PhiWsData is available
    ######################################################
    logger.debug(f"Checking for a workspace at {ws_root_path}")
    ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_path(ws_root_path)
    if ws_data is None:
        # This happens if
        # - The user is setting up an existing workspace
        # - the user ran `phi init -r` which erases the self._path_to_ws_data_map
        logger.debug(f"Could not find an existing workspace at path: {ws_root_path}")

        # In this case, the local workspace directory exists but PhiConf does not have a record
        print_info(f"Adding {ws_root_path} as a workspace")
        phi_conf.add_new_ws_to_config(
            ws_name=ws_root_path.stem,
            ws_root_path=ws_root_path,
        )
        ws_data = phi_conf.get_ws_data_by_path(ws_root_path)
        # If the ws_data is still none it means the workspace has bad data in it
        if ws_data is None:
            logger.error(f"Could not add workspace from: {ws_root_path}")
            logger.error("Please try again")
            return False
    else:
        logger.debug(f"Found workspace {ws_data.ws_name}")
        phi_conf.refresh_ws_config(ws_data.ws_name)

    ######################################################
    # 1.4 Set workspace as active
    ######################################################
    is_active_ws = True
    phi_conf.active_ws_name = ws_data.ws_name

    ######################################################
    # 1.5 Check if remote origin is available
    ######################################################
    _remote_origin_url: Optional[str] = get_remote_origin_for_dir(ws_root_path)
    logger.debug("Git origin: {}".format(_remote_origin_url))

    ######################################################
    ## 2. Create or Update WorkspaceSchema
    ######################################################
    # If a ws_schema exists for this workspace, this workspace has a record in the backend
    ws_schema: Optional[WorkspaceSchema] = ws_data.ws_schema

    ######################################################
    # 2.1 Create WorkspaceSchema for a NEWLY CREATED WORKSPACE
    ######################################################
    if ws_schema is None:
        from phiterm.api.workspace import create_workspace

        # If ws_schema is None, this is a NEWLY CREATED WORKSPACE.
        # We make a call to the backend to create a new ws_schema
        logger.debug("ws_schema not found, trying to create one... ")
        logger.debug("ws_name: {}".format(ws_data.ws_name))
        logger.debug("is_active_ws: {}".format(is_active_ws))

        if phi_conf.user is not None:
            ws_schema = create_workspace(
                user=phi_conf.user,
                workspace=WorkspaceSchema(
                    ws_name=ws_data.ws_name,
                    is_primary_ws_for_user=is_active_ws,
                ),
            )
        if ws_schema is not None:
            phi_conf.update_ws_data(ws_name=ws_data.ws_name, ws_schema=ws_schema)
    else:
        from phiterm.api.workspace import update_workspace

        # 2.2 Update WorkspaceSchema
        logger.debug("ws_schema found, updating... ")
        logger.debug("ws_name: {}".format(ws_data.ws_name))
        logger.debug("is_active_ws: {}".format(is_active_ws))

        if phi_conf.user is not None:
            ws_schema.is_primary_ws_for_user = is_active_ws
            ws_schema_updated = update_workspace(
                user=phi_conf.user,
                workspace=ws_schema,
            )
            if ws_schema_updated is not None:
                # Update the ws_schema for this workspace.
                phi_conf.update_ws_data(
                    ws_name=ws_data.ws_name, ws_schema=ws_schema_updated
                )

    ######################################################
    # 2.3 Refresh PhiWsData
    ######################################################
    # Refresh ws_data because phi_conf.update_ws_data()
    # will create a new ws_data object in PhiConf
    # Also, cast it to PhiWsData because ws_data is no longer Optional[PhiWsData]
    ws_data = cast(PhiWsData, phi_conf.get_ws_data_by_name(ws_data.ws_name))

    ######################################################
    ## 3. Complete Workspace setup
    ######################################################

    ######################################################
    # 3.1 Validate PhiWsData
    ######################################################
    ws_is_valid, pending_actions = phi_conf.validate_workspace(ws_name=ws_data.ws_name)

    if ws_is_valid and ws_data.ws_config is not None:
        scripts_dir = ws_data.ws_config.scripts_dir
        install_deps_file = f"sh {ws_root_path}/{scripts_dir}/install.sh"
        print_subheading(f"Setup complete! Next steps:\n")
        print_info("1. Deploy workspace:")
        print_info("\tphi ws up")
        print_info("2. Install workspace dependencies:")
        print_info(f"\t{install_deps_file}")
        # if pending_actions and len(pending_actions) > 0:
        #     print_info("3. Complete pending actions:")
        #     print_howtofix_pending_actions(pending_actions)
        return True
    else:
        print_info(f"Workspace setup pending.")
        print_howtofix_pending_actions(pending_actions)
        return False

    ######################################################
    ## End Workspace setup
    ######################################################


def filter_and_prep_configs(
    ws_config: WorkspaceConfig,
    target_env: Optional[str] = None,
    target_config: Optional[WorkspaceConfigType] = None,
    order: Optional[str] = "create",
) -> List[InfraConfig]:
    from phidata.utils.prep_infra_config import prep_infra_config

    # Step 1. Filter configs
    # 1.1: Filter by config type: docker/k8s/aws
    filtered_configs: List[InfraConfig] = []
    if target_config is None:
        if ws_config.docker is not None:
            filtered_configs.extend(ws_config.docker)
        if order == "delete":
            if ws_config.k8s is not None:
                filtered_configs.extend(ws_config.k8s)
            if ws_config.aws is not None:
                filtered_configs.extend(ws_config.aws)
        else:
            if ws_config.aws is not None:
                filtered_configs.extend(ws_config.aws)
            if ws_config.k8s is not None:
                filtered_configs.extend(ws_config.k8s)
    elif target_config == WorkspaceConfigType.docker:
        if ws_config.docker is not None:
            filtered_configs.extend(ws_config.docker)
        else:
            logger.error("No DockerConfig provided")
    elif target_config == WorkspaceConfigType.k8s:
        if ws_config.k8s is not None:
            filtered_configs.extend(ws_config.k8s)
        else:
            logger.error("No K8sConfig provided")
    elif target_config == WorkspaceConfigType.aws:
        if ws_config.aws is not None:
            filtered_configs.extend(ws_config.aws)
        else:
            logger.error("No AwsConfig provided")

    # 1.2: Filter by env: dev/stg/prd
    configs_after_env_filter: List[InfraConfig] = []
    if target_env is None or target_env == "all":
        configs_after_env_filter = filtered_configs
    else:
        for config in filtered_configs:
            if config.env is None:
                continue
            if target_env == config.env:
                configs_after_env_filter.append(config)
    # logger.debug("Filtered configs: {}`".format(configs_after_env_filter))

    # Step 2. Prepare configs
    ready_to_use_configs: List[InfraConfig] = []
    for config in configs_after_env_filter:
        if not isinstance(config, InfraConfig):
            logger.error(f"{config} is not an instance of InfraConfig")
            continue
        if not config.is_valid():
            logger.warning(f"Skipping {config.__class__}\n")
            continue
        if not config.enabled:
            logger.debug(
                f"{config.__class__.__name__} for env {config.__class__.env} disabled"
            )
            continue

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        ######################################################################

        _config = prep_infra_config(
            infra_config=config,
            ws_config=ws_config,
        )
        ready_to_use_configs.append(_config)

    # Return filtered and prepared configs
    return ready_to_use_configs


def deploy_workspace(
    ws_data: PhiWsData,
    target_env: Optional[str] = None,
    target_config: Optional[WorkspaceConfigType] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    target_app: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> None:
    """Deploy a Phidata Workspace. This is called from `phi ws up`"""
    from phidata.docker.config import DockerConfig
    from phidata.k8s.config import K8sConfig
    from phidata.aws.config import AwsConfig

    if ws_data is None is None:
        logger.error("WorkspaceConfig invalid")
        return

    ws_config: WorkspaceConfig = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    configs_to_deploy: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
        target_config=target_config,
        order="create",
    )

    num_configs_to_deploy = len(configs_to_deploy)
    num_configs_deployed = 0
    for config in configs_to_deploy:
        logger.debug(f"Deploying {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phiterm.docker.docker_operator import deploy_docker_config

            deploy_docker_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_deployed += 1
        if isinstance(config, K8sConfig):
            from phiterm.k8s.k8s_operator import deploy_k8s_config

            deploy_k8s_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_deployed += 1
        if isinstance(config, AwsConfig):
            from phiterm.aws.aws_operator import deploy_aws_config

            deploy_aws_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_deployed += 1
        # white space between runs
        print_info("")

    print_info(f"# Configs deployed: {num_configs_deployed}/{num_configs_to_deploy}\n")
    if num_configs_to_deploy == num_configs_deployed:
        if not dry_run:
            print_subheading("Workspace deploy success")
    else:
        logger.error("Workspace deploy failed")


def shutdown_workspace(
    ws_data: PhiWsData,
    target_env: Optional[str] = None,
    target_config: Optional[WorkspaceConfigType] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    target_app: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> None:
    """Shutdown the Phidata Workspace. This is called from `phi ws down`"""
    from phidata.docker.config import DockerConfig
    from phidata.k8s.config import K8sConfig
    from phidata.aws.config import AwsConfig

    if ws_data is None is None:
        logger.error("WorkspaceConfig invalid")
        return

    ws_config: WorkspaceConfig = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    configs_to_shutdown: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
        target_config=target_config,
        order="delete",
    )

    num_configs_to_shutdown = len(configs_to_shutdown)
    num_configs_shutdown = 0
    for config in configs_to_shutdown:
        logger.debug(f"Shutting down {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phiterm.docker.docker_operator import shutdown_docker_config

            shutdown_docker_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_shutdown += 1
        if isinstance(config, K8sConfig):
            from phiterm.k8s.k8s_operator import shutdown_k8s_config

            shutdown_k8s_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_shutdown += 1
        if isinstance(config, AwsConfig):
            from phiterm.aws.aws_operator import shutdown_aws_config

            shutdown_aws_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_shutdown += 1
        # white space between runs
        print_info("")

    print_info(
        f"\n# Configs shutdown: {num_configs_shutdown}/{num_configs_to_shutdown}"
    )
    if num_configs_to_shutdown == num_configs_shutdown:
        if not dry_run:
            print_subheading("Workspace shutdown success")
    else:
        print_subheading("Workspace shutdown failed")


def patch_workspace(
    ws_data: PhiWsData,
    target_env: Optional[str] = None,
    target_config: Optional[WorkspaceConfigType] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    target_app: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> None:
    """Patch the Phidata Workspace. This is called from `phi ws patch`"""
    from phidata.docker.config import DockerConfig
    from phidata.k8s.config import K8sConfig
    from phidata.aws.config import AwsConfig

    if ws_data is None is None:
        logger.error("WorkspaceConfig invalid")
        return

    ws_config: WorkspaceConfig = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    configs_to_patch: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
        target_config=target_config,
        order="create",
    )

    num_configs_to_patch = len(configs_to_patch)
    num_configs_patched = 0
    for config in configs_to_patch:
        logger.debug(f"Patching {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phiterm.docker.docker_operator import patch_docker_config

            patch_docker_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_patched += 1
        if isinstance(config, K8sConfig):
            from phiterm.k8s.k8s_operator import patch_k8s_config

            patch_k8s_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_patched += 1
        if isinstance(config, AwsConfig):
            from phiterm.aws.aws_operator import patch_aws_config

            patch_aws_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                app_filter=target_app,
                dry_run=dry_run,
                auto_confirm=auto_confirm,
            )
            num_configs_patched += 1
        # white space between runs
        print_info("")

    print_info(f"\n# Configs patched: {num_configs_patched}/{num_configs_to_patch}")
    if num_configs_to_patch == num_configs_patched:
        if not dry_run:
            print_subheading("Workspace patch success")
    else:
        print_subheading("Workspace patch failed")


def set_workspace_as_active(ws_name: Optional[str], refresh: bool = True) -> bool:
    from phiterm.api.workspace import update_primary_workspace

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check PhiConf is valid
    ######################################################
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    ######################################################
    # 1.2 Check ws_root_path is valid
    ######################################################
    # By default, we assume this command is run from the workspace directory
    ws_root_path: Optional[Path] = None
    if ws_name is None:
        # If the user does not provide a ws_name, that implies `phi set` is ran from
        # the workspace directory.
        ws_root_path = Path(".").resolve()
    else:
        # If the user provides a workspace name manually, we find the dir for that ws
        ws_root_path = phi_conf.get_ws_root_path_by_name(ws_name)
        if ws_root_path is None:
            logger.error(f"Could not find workspace {ws_name}")
            return False

    ws_dir_is_valid: bool = (
        ws_root_path is not None and ws_root_path.exists() and ws_root_path.is_dir()
    )
    if not ws_dir_is_valid:
        logger.error("Invalid workspace directory: {}".format(ws_root_path))
        return False

    ######################################################
    # 1.3 Validate PhiWsData is available i.e. a workspace is available at this directory
    ######################################################
    logger.debug(f"Checking for a workspace at path: {ws_root_path}")
    active_ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_path(ws_root_path)
    if active_ws_data is None:
        # This happens when the workspace is not yet setup
        print_info(f"Could not find a workspace at path: {ws_root_path}")
        print_info(
            f"If this workspace has not been setup, please run `phi ws setup` from the workspace directory"
        )
        return False

    new_active_ws_name: str = active_ws_data.ws_name
    print_heading(f"Setting workspace {new_active_ws_name} as active")
    if refresh:
        try:
            phi_conf.refresh_ws_config(new_active_ws_name)
        except WorkspaceException as e:
            logger.error(
                "Could not refresh workspace config, please fix errors and try again"
            )
            logger.error(e)
            return False

    ######################################################
    # 1.4 Make api request if updating active workspace
    ######################################################
    logger.debug("Updating active workspace api")

    if phi_conf.user is not None:
        ws_schema: Optional[WorkspaceSchema] = active_ws_data.ws_schema
        if ws_schema is None:
            logger.warning(
                f"Please setup {new_active_ws_name} by running `phi ws setup`"
            )
        else:
            updated_workspace_schema = update_primary_workspace(
                user=phi_conf.user,
                workspace=ws_schema,
            )
            if updated_workspace_schema is not None:
                # Update the ws_schema for this workspace.
                phi_conf.update_ws_data(
                    ws_name=new_active_ws_name, ws_schema=updated_workspace_schema
                )

    ######################################################
    ## 2. Set workspace as active
    ######################################################
    phi_conf.active_ws_name = new_active_ws_name
    print_info("Active workspace updated")
    return True


def delete_workspace(ws_to_delete: Optional[List[str]], phi_conf: PhiConf) -> bool:
    from phiterm.api.workspace import delete_workspaces_api

    if ws_to_delete is None or len(ws_to_delete) == 0:
        return True

    logger.debug(f"Deleting workspaces: {ws_to_delete}")
    workspaces_deleted = delete_workspaces_api(
        user=phi_conf.user, workspaces_to_delete=ws_to_delete
    )
    if workspaces_deleted:
        pass
        # phi_conf.delete_ws_data(ws_name=ws)
    return workspaces_deleted


def print_workspace_status(
    ws_data: PhiWsData,
    target_env: Optional[str] = None,
    target_config: Optional[WorkspaceConfigType] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    target_app: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> None:
    """Print the Workspace status. This is called from `phi ws status`"""
    from phidata.docker.config import DockerConfig
    from phidata.k8s.config import K8sConfig
    from phidata.aws.config import AwsConfig

    if ws_data is None is None:
        logger.error("WorkspaceConfig invalid")
        return

    ws_config: WorkspaceConfig = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    configs_to_patch: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
        target_config=target_config,
        order="create",
    )

    num_configs_to_patch = len(configs_to_patch)
    num_configs_patched = 0
    for config in configs_to_patch:
        logger.debug(f"Patching {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phiterm.docker.docker_operator import patch_docker_config

            # print_docker_config_status(
            #     config=config,
            #     name_filter=target_name,
            #     type_filter=target_type,
            #     app_filter=target_app,
            #     dry_run=dry_run,
            #     auto_confirm=auto_confirm,
            # )
            num_configs_patched += 1
        if isinstance(config, K8sConfig):
            from phiterm.k8s.k8s_operator import patch_k8s_config

            # print_k8s_config_status(
            #     config=config,
            #     name_filter=target_name,
            #     type_filter=target_type,
            #     app_filter=target_app,
            #     dry_run=dry_run,
            #     auto_confirm=auto_confirm,
            # )
            num_configs_patched += 1
        if isinstance(config, AwsConfig):
            from phiterm.aws.aws_operator import patch_aws_config

            # print_aws_config_status(
            #     config=config,
            #     name_filter=target_name,
            #     type_filter=target_type,
            #     app_filter=target_app,
            #     dry_run=dry_run,
            #     auto_confirm=auto_confirm,
            # )
            num_configs_patched += 1
        # white space between runs
        print_info("")


######################################################
## Deprecated functions
######################################################
#
# def print_git_setup(
#     ws_name: str,
#     ws_root_path: Path,
#     version_control_provider: Optional[
#         VersionControlProviderEnum
#     ] = VersionControlProviderEnum.GITHUB,
# ) -> None:
#     """Prints commands to setup a remote git repo for a workspace"""
#     from phiterm.utils.git import get_remote_origin_for_dir
#
#     # Check if a remote origin is available
#     _remote_origin_url: Optional[str] = get_remote_origin_for_dir(ws_root_path)
#     # logger.debug("_remote_origin_url: {}".format(_remote_origin_url))
#     if _remote_origin_url:
#         print_info(f"Remote origin already set: {_remote_origin_url}")
#         return
#
#     create_new_git_repo_url = "https://github.com/new"
#     create_remote_repo_url: str
#     # TODO: Add more repo urls
#     if version_control_provider == VersionControlProviderEnum.GITHUB:
#         create_remote_repo_url = create_new_git_repo_url
#     else:
#         create_remote_repo_url = create_new_git_repo_url
#
#     print_subheading(f"\nSteps to setup a git repo for workspace: {ws_name}")
#     print_info(f"\nCreate a new git repo named {ws_name} at {create_remote_repo_url}")
#     print_info(
#         f"Then follow the following steps to initialize the directory {ws_name} as a git repo and run git push"
#     )
#     print_info(
#         f"NOTE: These steps will be provided by your version control provider (github, gitlab..), but would look something like:"
#     )
#     print_info(f"       git init")
#     print_info(f"       git add .")
#     print_info(f"       git commit -m 'Init Phidata Workspace'")
#     print_info(f"       git branch -M main")
#     print_info(
#         f"       git remote add origin https://github.com/<USERNAME>/{ws_name}.git"
#     )
#     print_info(f"       git push -u origin main")
#     print_info(f"Run `phi ws setup` after running git push to update your workspace\n")
#
#     _open_browser = typer.confirm(
#         f"Would you like to open your browser to visit: {create_remote_repo_url}\n-->"
#     )
#     if _open_browser:
#         typer.launch(create_remote_repo_url)
#
#
# def clone_workspace(ws_to_clone: WorkspaceSchema, phi_conf: PhiConf) -> None:
#     """Clones an available phidata workspace on the users machine. To clone a ws, the PhiConf
#     already has the ws_data and ws_schema available. We then update the
#     ws_root_path and ws_config_file_path in the ws_data. After updating the
#     path, we're call setup_workspace() to go through our usual setup flow
#     """
#     from phiterm.utils.git import GitCloneProgress
#     from phiterm.workspace.ws_utils import get_ws_config_file_path
#
#     # Check if we can create the workspace in the current dir
#     current_dir: Path = Path("../../workspace").resolve()
#     logger.debug(f"current_dir: {current_dir}")
#     ws_root_path: Path = current_dir.joinpath(ws_to_clone.name)
#     if ws_root_path.exists():
#         logger.error(
#             f"{ws_root_path} already exists, please choose another directory to clone or delete the existing dir and run `phi ws init` again"
#         )
#         return
#
#     print_info(f"Creating {str(ws_root_path)}")
#     # Clone the workspace repo in ws_root_path
#     repo_to_clone = ws_to_clone.git_url
#     _cloned_git_repo: git.Repo = git.Repo.clone_from(
#         repo_to_clone, str(ws_root_path), progress=GitCloneProgress()  # type: ignore
#     )
#     # Validate ws_root_path exists and is a directory
#     if not (ws_root_path.exists() and ws_root_path.is_dir()):
#         logger.error(f"Could not clone workspace, please try again")
#         return
#     # Get ws_config_file_path and validate
#     ws_config_file_path: Path = get_ws_config_file_path(ws_root_path)
#     if not (ws_config_file_path.exists() and ws_config_file_path.is_file()):
#         logger.error(f"Could not find workspace config at: {ws_config_file_path}")
#         return
#     # Update the ws_data in PhiConf
#     phi_conf.update_ws_data(
#         ws_name=ws_to_clone.ws_name,
#         ws_root_path=ws_root_path,
#         ws_config_file_path=ws_config_file_path,
#     )
#     setup_workspace(ws_root_path)
#
#
# def init_workspace_data(ws_schema: WorkspaceSchema, phi_conf: PhiConf):
#     """Initializes workspace data like cloud projects"""
#
#     logger.debug("*=* Initializing workspace data for: {}".format(ws_schema.name))
#     logger.debug("TO_BE_IMPLEMENTED")
#     logger.debug("*=* Workspace data init complete: {}".format(ws_schema.name))
#
# def import_workspace():
#     # TODO: use code to clone/import existing workspace later
#     available_ws = phi_conf.available_ws
#     if available_ws:
#         print_info("Would you like to:")
#         print_info("  [1] Create a new workspace")
#         print_info("  [2] Clone an existing workspace")
#         print_info("  [3] Enter any other key to exit")
#         ws_inp_raw = input("--> ")
#         ws_inp_int = str_to_int(ws_inp_raw)
#         if ws_inp_int == 1:
#             create_workspace_using_input()
#         elif ws_inp_int == 2:
#             print_info("\nSelect workspace to clone:")
#             for idx, avl_ws in enumerate(available_ws, start=1):
#                 print_info("  [{}] {}".format(idx, avl_ws.name))
#             clone_ws_inp_raw = input("--> ")
#             clone_ws_inp_int = str_to_int(clone_ws_inp_raw)
#             if clone_ws_inp_int:
#                 ws_to_clone = available_ws[clone_ws_inp_int - 1]
#                 clone_workspace(ws_to_clone, phi_conf)
#             return
#         return
#     else:
#



from pathlib import Path
from typing import List, Optional, cast, Dict, Tuple, Any

from phidata.infra.config import InfraConfig
from phidata.k8s.api_client import K8sApiClient
from phidata.k8s.config import K8sConfig
from phidata.k8s.exceptions import K8sConfigException
from phidata.k8s.manager import K8sManager
from phidata.k8s.resource.types import K8sResourceType, K8sResource
from phidata.k8s.resource.group import K8sResourceGroup
from phidata.k8s.resource.utils import get_k8s_resources_from_group
from phidata.product import DataProduct
from phidata.workflow import Workflow
from phidata.types.context import PathContext, RunContext

from phiterm.workspace.phi_ws_data import PhiWsData
from phiterm.workspace.ws_enums import WorkspaceConfigType
from phiterm.workspace.ws_operator import filter_and_prep_configs
from phiterm.workspace.ws_utils import get_ws_config_dir_path
from phiterm.types.run_status import RunStatus
from phiterm.utils.filesystem import delete_files_in_dir
from phiterm.utils.cli_console import (
    print_error,
    print_heading,
    print_subheading,
    print_info,
    print_warning,
)
from phiterm.utils.log import logger


def deploy_k8s_config(
    config: K8sConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")
    print_heading(f"--**-- K8s env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        k8s_manager.create_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Create resources
    try:
        success: bool = k8s_manager.create_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        raise

    # Step 4: Validate resources are created
    resource_creation_valid: bool = k8s_manager.validate_resources_are_created(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resource_creation_valid:
        logger.error("K8sResource creation could not be validated")
        return False

    print_info("K8s config deployed")
    return True


def shutdown_k8s_config(
    config: K8sConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")
    print_heading(f"--**-- K8s env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        k8s_manager.delete_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Delete resources
    try:
        success: bool = k8s_manager.delete_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        return False

    # Step 4: Validate resources are deleted
    resources_deletion_valid: bool = k8s_manager.validate_resources_are_deleted(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resources_deletion_valid:
        logger.error("K8sResource deletion could not be validated")
        return False

    print_info("K8s config shut down")
    return True


def patch_k8s_config(
    config: K8sConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")
    print_heading(f"--**-- K8s env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        k8s_manager.patch_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Patch resources
    try:
        success: bool = k8s_manager.patch_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        raise

    # Step 4: Validate resources are patched
    resources_patch_valid: bool = k8s_manager.validate_resources_are_patched(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resources_patch_valid:
        logger.error("K8sResource patch could not be validated")
        return False

    print_info("K8s config patched")
    return True


def save_k8s_resources(
    ws_data: PhiWsData,
    target_env: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    target_app: Optional[str] = None,
) -> None:
    """Saves the K8s resources"""

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return
    ws_config = ws_data.ws_config
    # Set the local environment variables before processing configs
    ws_config.set_local_env()

    configs_to_save: List[InfraConfig] = filter_and_prep_configs(
        ws_config=ws_config,
        target_env=target_env,
        target_config=WorkspaceConfigType.k8s,
        order="create",
    )

    num_configs_to_save = len(configs_to_save)
    num_configs_saved = 0
    for config in configs_to_save:
        if not isinstance(config, K8sConfig):
            continue

        env = config.env
        print_heading(
            "\nBuilding K8s manifests{}\n".format(
                f" for env: {env}" if env is not None else ""
            )
        )

        # Step 2: Get the K8sManager
        k8s_manager: K8sManager = config.get_k8s_manager()
        if k8s_manager is None:
            raise K8sConfigException("K8sManager unavailable")

        # Step 3: Prep the directory to write to
        ws_root_path: Optional[Path] = ws_data.ws_root_path
        if ws_root_path is None:
            print_error("ws_root_path invalid")
            continue
        workspace_config_dir: Optional[Path] = get_ws_config_dir_path(ws_root_path)
        if workspace_config_dir is None:
            print_error("workspace_config_dir invalid")
            continue

        resources_dir: Path = workspace_config_dir.joinpath(config.resources_dir)
        if config.env is not None:
            resources_dir = resources_dir.joinpath(config.env)
        if config.args.name is not None:
            resources_dir = resources_dir.joinpath(config.args.name)

        # Step 4: Get the K8sResources
        k8s_resources: Optional[
            Dict[str, K8sResourceGroup]
        ] = k8s_manager.get_resource_groups(app_filter=target_app)
        if k8s_resources is None or len(k8s_resources) == 0:
            print_error("No K8sResources to save")
            return

        # delete directory to save resources if needed
        if resources_dir.exists():
            print_info(f"Deleting {str(resources_dir)}")
            if resources_dir.is_file():
                resources_dir.unlink()
            elif resources_dir.is_dir():
                delete_files_in_dir(resources_dir)
        print_info(f"Saving to {str(resources_dir)}")
        resources_dir.mkdir(exist_ok=True, parents=True)

        for rg_name, resource_group in k8s_resources.items():
            print_info(f"Processing {rg_name}")
            rg_dir: Path = resources_dir.joinpath(rg_name)
            rg_dir.mkdir(exist_ok=True)

            resource_list: List[K8sResource] = get_k8s_resources_from_group(
                k8s_resource_group=resource_group,
                name_filter=target_name,
                type_filter=target_type,
            )

            if resource_list is None:
                continue

            for resource in resource_list:
                if resource is not None:
                    resource_name = resource.get_resource_name()
                    resource_file: Path = rg_dir.joinpath(f"{resource_name}.yaml")
                    try:
                        manifest_yaml = resource.get_k8s_manifest_yaml(
                            default_flow_style=False
                        )  # use default_style='"' for debugging
                        if manifest_yaml is not None:
                            logger.debug(f"Writing {str(resource_file)}")
                            resource_file.write_text(manifest_yaml)
                    except Exception as e:
                        logger.error(f"Could not parse {resource_name}: {e}")
                        continue
        num_configs_saved += 1

    print_info(f"\n# Configs built: {num_configs_saved}/{num_configs_to_save}")


def run_workflows_k8s(
    workflow_file: str,
    workflows: Dict[str, Workflow],
    run_context: RunContext,
    k8s_config: K8sConfig,
    target_app: Optional[str] = None,
    workflow_name: Optional[str] = None,
    use_dag_id: Optional[str] = None,
) -> bool:
    from phidata.app.databox import Databox, DataboxArgs, default_databox_name
    from phidata.k8s.enums.api_version import ApiVersion
    from phidata.k8s.enums.kind import Kind
    from phidata.k8s.resource.types import K8sResourceType
    from phidata.k8s.resource.apps.v1.deployment import Deployment
    from phidata.k8s.resource.core.v1.pod import Pod
    from phidata.k8s.resource.meta.v1.object_meta import ObjectMeta

    logger.debug("Running Workflow in K8sContainer")
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = k8s_config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")

    # Step 2: Check if a Databox is available for running the Workflow
    # If available get the DataboxArgs
    databox_app_name = target_app or k8s_config.databox_name or default_databox_name
    logger.debug(f"Using App: {databox_app_name}")
    databox_app = k8s_config.get_app_by_name(databox_app_name)
    if databox_app is None or not isinstance(databox_app, Databox):
        print_error("Databox not available")
        return False
    databox_app_args: DataboxArgs = databox_app.args
    # logger.debug(f"DataboxArgs: {databox_app_args}")
    if databox_app_args is None or not isinstance(databox_app_args, DataboxArgs):
        print_error("DataboxArgs invalid")
        return False
    databox_app_args = cast(DataboxArgs, databox_app_args)

    # Step 3: Build the PathContext for the workflows.
    # NOTE: The PathContext uses directories relative to the
    # workspace_parent_container_path
    workspace_name = k8s_config.workspace_root_path.stem
    workspace_root_container_path = Path(
        databox_app_args.workspace_parent_container_path
    ).joinpath(workspace_name)
    scripts_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.scripts_dir
    )
    storage_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.storage_dir
    )
    meta_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.meta_dir
    )
    products_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.products_dir
    )
    notebooks_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.notebooks_dir
    )
    workspace_config_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.workspace_config_dir
    )
    workflow_file_path = Path(products_dir_container_path).joinpath(workflow_file)
    wf_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_container_path,
        storage_dir=storage_dir_container_path,
        meta_dir=meta_dir_container_path,
        products_dir=products_dir_container_path,
        notebooks_dir=notebooks_dir_container_path,
        workspace_config_dir=workspace_config_dir_container_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {wf_path_context.json(indent=4)}")

    # Step 4: Gather Pod and K8sApiClient to use
    # Step 4.1: Get the deployment to run the Workflow
    databox_deployments: Optional[List[K8sResourceType]] = k8s_manager.read_resources(
        name_filter=databox_app_name, type_filter="Deployment"
    )
    # logger.debug(f"databox_deployments: {databox_deployments}")
    if databox_deployments is None or len(databox_deployments) == 0:
        logger.error(f"Deployment: {databox_app_name} not found")
        return False
    if len(databox_deployments) > 1:
        logger.info(
            f"Found {len(databox_deployments)} deployments for {databox_app_name}, using the first one."
        )
    databox_deploy: Deployment = databox_deployments[0]
    # Step 4.2: Get the deployment name which generates the pod name
    databox_deploy_name = databox_deploy.get_resource_name()
    if databox_deploy_name is None:
        logger.error(f"Deployment name not available")
        return False
    logger.debug("databox_deploy_name: {}".format(databox_deploy_name))
    # Step 4.3: Get the pod using the databox_deploy_name
    databox_pod = Pod(
        name=databox_deploy_name,
        api_version=ApiVersion.CORE_V1,
        kind=Kind.POD,
        metadata=ObjectMeta(
            name=databox_deploy_name,
            namespace=k8s_config.namespace,
        ),
    )
    # Step 4.4: Get the container name
    databox_container_name = databox_app.get_container_name()
    # Step 4.5: Get the K8sApiClient
    k8s_api_client: K8sApiClient = k8s_manager.k8s_worker.k8s_api_client

    # Step 5: Run single Workflow if workflow_name is provided
    if workflow_name is not None:
        if workflow_name not in workflows:
            print_error(
                "Could not find '{}' in {}".format(
                    workflow_name, "[{}]".format(", ".join(workflows.keys()))
                )
            )
            return False

        wf_run_success: bool = False
        workflow_to_run = workflows[workflow_name]
        _name = workflow_name or workflow_to_run.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        workflow_to_run.run_context = run_context
        workflow_to_run.path_context = wf_path_context
        # Use DataProduct dag_id if provided
        if use_dag_id is not None:
            workflow_to_run.dag_id = use_dag_id
        wf_run_success = workflow_to_run.run_in_k8s_container(
            pod=databox_pod,
            k8s_api_client=k8s_api_client,
            container_name=databox_container_name,
            k8s_env=k8s_config.k8s_env,
        )

        print_subheading("\nWorkflow run status:")
        print_info("{}: {}".format(_name, "Success" if wf_run_success else "Fail"))
        print_info("")
        return wf_run_success
    # Step 6: Run all Workflows if workflow_name is None
    else:
        wf_run_status: List[RunStatus] = []
        for wf_name, wf_obj in workflows.items():
            _name = wf_name or wf_obj.name
            print_subheading(f"\nRunning {_name}")
            # Pass down context
            wf_obj.run_context = run_context
            wf_obj.path_context = wf_path_context
            run_success = wf_obj.run_in_k8s_container(
                pod=databox_pod,
                k8s_api_client=k8s_api_client,
                container_name=databox_container_name,
                k8s_env=k8s_config.k8s_env,
            )
            wf_run_status.append(RunStatus(name=_name, success=run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True


def run_data_products_k8s(
    workflow_file: str,
    data_products: Dict[str, DataProduct],
    run_context: RunContext,
    k8s_config: K8sConfig,
    target_app: Optional[str] = None,
) -> bool:
    from phidata.app.databox import Databox, DataboxArgs, default_databox_name
    from phidata.k8s.enums.api_version import ApiVersion
    from phidata.k8s.enums.kind import Kind
    from phidata.k8s.resource.types import K8sResourceType
    from phidata.k8s.resource.apps.v1.deployment import Deployment
    from phidata.k8s.resource.core.v1.pod import Pod
    from phidata.k8s.resource.meta.v1.object_meta import ObjectMeta

    logger.debug("Running DataProducts in K8sContainer")
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = k8s_config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")

    # Step 2: Check if a Databox is available for running the Workflow
    # If available get the DataboxArgs
    databox_app_name = target_app or k8s_config.databox_name or default_databox_name
    logger.debug(f"Using App: {databox_app_name}")
    databox_app = k8s_config.get_app_by_name(databox_app_name)
    if databox_app is None or not isinstance(databox_app, Databox):
        print_error("Databox not available")
        return False
    databox_app_args: DataboxArgs = databox_app.args
    # logger.debug(f"DataboxArgs: {databox_app_args}")
    if databox_app_args is None or not isinstance(databox_app_args, DataboxArgs):
        print_error("DataboxArgs invalid")
        return False
    databox_app_args = cast(DataboxArgs, databox_app_args)

    # Step 3: Build the PathContext for the DataProducts.
    # NOTE: The PathContext uses directories relative to the
    # workspace_parent_container_path
    workspace_name = k8s_config.workspace_root_path.stem
    workspace_root_container_path = Path(
        databox_app_args.workspace_parent_container_path
    ).joinpath(workspace_name)
    scripts_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.scripts_dir
    )
    storage_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.storage_dir
    )
    meta_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.meta_dir
    )
    products_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.products_dir
    )
    notebooks_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.notebooks_dir
    )
    workspace_config_dir_container_path = workspace_root_container_path.joinpath(
        k8s_config.workspace_config_dir
    )
    workflow_file_path = Path(products_dir_container_path).joinpath(workflow_file)
    dp_path_context: PathContext = PathContext(
        scripts_dir=scripts_dir_container_path,
        storage_dir=storage_dir_container_path,
        meta_dir=meta_dir_container_path,
        products_dir=products_dir_container_path,
        notebooks_dir=notebooks_dir_container_path,
        workspace_config_dir=workspace_config_dir_container_path,
        workflow_file=workflow_file_path,
    )
    logger.debug(f"PathContext: {dp_path_context.json(indent=4)}")

    # Step 4: Gather Pod and K8sApiClient to use
    # Step 4.1: Get the deployment to run the Workflow
    databox_deployments: Optional[List[K8sResourceType]] = k8s_manager.read_resources(
        name_filter=databox_app_name, type_filter="Deployment"
    )
    # logger.debug(f"databox_deployments: {databox_deployments}")
    if databox_deployments is None or len(databox_deployments) == 0:
        logger.error(f"Deployment: {databox_app_name} not found")
        return False
    if len(databox_deployments) > 1:
        logger.info(
            f"Found {len(databox_deployments)} deployments for {databox_app_name}, using the first one."
        )
    databox_deploy: Deployment = databox_deployments[0]
    # Step 4.2: Get the deployment name which generates the pod name
    databox_deploy_name = databox_deploy.get_resource_name()
    if databox_deploy_name is None:
        logger.error(f"Deployment name not available")
        return False
    logger.debug("databox_deploy_name: {}".format(databox_deploy_name))
    # Step 4.3: Get the pod using the databox_deploy_name
    databox_pod = Pod(
        name=databox_deploy_name,
        api_version=ApiVersion.CORE_V1,
        kind=Kind.POD,
        metadata=ObjectMeta(
            name=databox_deploy_name,
            namespace=k8s_config.namespace,
        ),
    )
    # Step 4.4: Get the container name
    databox_container_name = databox_app.get_container_name()
    # Step 4.5: Get the K8sApiClient
    k8s_api_client: K8sApiClient = k8s_manager.k8s_worker.k8s_api_client

    # Step 5: Run the DataProducts
    dp_run_status: List[RunStatus] = []
    for dp_name, dp_obj in data_products.items():
        _name = dp_name or dp_obj.name
        print_subheading(f"\nRunning {_name}")
        # Pass down context
        dp_obj.run_context = run_context
        dp_obj.path_context = dp_path_context
        run_success = dp_obj.run_in_k8s_container(
            pod=databox_pod,
            k8s_api_client=k8s_api_client,
            container_name=databox_container_name,
            k8s_env=k8s_config.k8s_env,
        )
        dp_run_status.append(RunStatus(name=_name, success=run_success))

    print_subheading("DataProduct run status:")
    print_info(
        "\n".join(
            [
                "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                for wf in dp_run_status
            ]
        )
    )
    print_info("")
    for _run in dp_run_status:
        if not _run.success:
            return False
    return True


def run_command_k8s(
    command: List[str],
    k8s_config: K8sConfig,
    target_app: Optional[str] = None,
) -> bool:
    from phidata.app.databox import Databox, DataboxArgs, default_databox_name
    from phidata.k8s.enums.api_version import ApiVersion
    from phidata.k8s.enums.kind import Kind
    from phidata.k8s.resource.types import K8sResourceType
    from phidata.k8s.resource.apps.v1.deployment import Deployment
    from phidata.k8s.resource.core.v1.pod import Pod
    from phidata.k8s.resource.meta.v1.object_meta import ObjectMeta
    from phidata.k8s.utils.pod import execute_command

    logger.debug("Running command in K8sContainer")
    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = k8s_config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")

    # Step 2: Check if a Databox is available for running the Workflow
    # If available get the DataboxArgs
    databox_app_name = target_app or k8s_config.databox_name or default_databox_name
    logger.debug(f"Using App: {databox_app_name}")
    databox_app = k8s_config.get_app_by_name(databox_app_name)
    if databox_app is None or not isinstance(databox_app, Databox):
        print_error("Databox not available")
        return False
    databox_app_args: DataboxArgs = databox_app.args
    # logger.debug(f"DataboxArgs: {databox_app_args}")
    if databox_app_args is None or not isinstance(databox_app_args, DataboxArgs):
        print_error("DataboxArgs invalid")
        return False
    databox_app_args = cast(DataboxArgs, databox_app_args)

    # Step 3: Gather Pod and K8sApiClient to use
    # Step 3.1: Get the deployment to run the Workflow
    databox_deployments: Optional[List[K8sResourceType]] = k8s_manager.read_resources(
        name_filter=databox_app_name, type_filter="Deployment"
    )
    # logger.debug(f"databox_deployments: {databox_deployments}")
    if databox_deployments is None or len(databox_deployments) == 0:
        logger.error(f"Deployment: {databox_app_name} not found")
        return False
    if len(databox_deployments) > 1:
        logger.info(
            f"Found {len(databox_deployments)} deployments for {databox_app_name}, using the first one."
        )
    databox_deploy: Deployment = databox_deployments[0]
    # Step 3.2: Get the deployment name which generates the pod name
    databox_deploy_name = databox_deploy.get_resource_name()
    if databox_deploy_name is None:
        logger.error(f"Deployment name not available")
        return False
    logger.debug("databox_deploy_name: {}".format(databox_deploy_name))
    # Step 3.3: Get the pod using the databox_deploy_name
    databox_pod = Pod(
        name=databox_deploy_name,
        api_version=ApiVersion.CORE_V1,
        kind=Kind.POD,
        metadata=ObjectMeta(
            name=databox_deploy_name,
            namespace=k8s_config.namespace,
        ),
    )
    # Step 3.4: Get the container name
    databox_container_name = databox_app.get_container_name()
    # Step 3.5: Get the K8sApiClient
    k8s_api_client: K8sApiClient = k8s_manager.k8s_worker.k8s_api_client

    # Step 4: Run the command
    print_subheading(f"Running command: `{command}`")
    run_status = execute_command(
        cmd=command,
        pod=databox_pod,
        k8s_api_client=k8s_api_client,
        container_name=databox_container_name,
        k8s_env=k8s_config.k8s_env,
    )
    logger.debug(f"Run status: {run_status}")
    return run_status



from pathlib import Path

APP_NAME = "phiterm"
APP_VERSION = "0.1.0"

DEFAULT_WS_NAME = "data-platform"

PHI_CONF_DIR: Path = Path.home().resolve().joinpath(".phi")
PHI_CONF_PATH: Path = PHI_CONF_DIR.joinpath("conf")
PHI_AUTH_TOKEN_PATH: Path = PHI_CONF_DIR.joinpath("token")
PHI_CREDS_PATH: Path = PHI_CONF_DIR.joinpath("creds")

# Rest API Constants
PHI_API_MODE = "prd"
PHI_AUTH_TOKEN_COOKIE: str = "__phi_session"
PHI_AUTH_TOKEN_HEADER: str = "X-PHIDATA-AUTH-TOKEN"
PHI_SIGNIN_URL_WITHOUT_PARAMS: str = (
    "http://localhost:3000/signin"
    if PHI_API_MODE == "dev"
    else "https://phidata.com/signin"
)
BACKEND_API_URL: str = (
    "http://localhost:8000" if PHI_API_MODE == "dev" else "https://api.phidata.com"
)

# Logger Names
PHI_LOGGER_NAME = "phi"
PHIDATA_LOGGER_NAME = "phidata"



# from pathlib import Path
# from typing import Any, Dict, List, Optional
#
# import pydantic
# import yaml
#
# import pak8.enums as pak8_enums
# from pak8.conf import Pak8Conf
# from pak8.gcp.gcp_pak8 import GCPPak8
# from pak8.k8s.pak8_k8s_client import Pak8K8sClient
# from pak8.k8s.resources.kubeconfig import Kubeconfig
# from phi import schemas
# from phiterm.conf.phi_conf import PhiConf, PhiK8sData
# from phiterm.k8s.k8s_ops import (
#     apply_k8s_resources_for_k8s_client,
#     get_active_k8s_resources_for_k8s_client,
#     get_k8s_resource_groups_as_dicts,
#     get_k8s_resources_as_dicts,
# )
# from phiterm.utils.common import pprint_error, pprint_info, pprint_status, pprint_subheading
# from phiterm.utils.filesystem import delete_files_in_dir
# from phiterm.utils.log import logger
#
#
# def save_k8s_resources_as_yaml(
#     ws_schema: schemas.WorkspaceSchema,
#     config: PhiConf,
#     refresh: bool = False,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> bool:
#     """Saves the K8s resources for a workspace in the `phi` dir for that workspace.
#     There are 2 ways of generating the K8s resources:
#     1. API call to zeus, zeus sends back the manifests as dicts, we save them as yaml
#     2. Use Pak8 directly, this leads to a faster development process
#
#     Currently we're using 2, but switching to 1 is easy if needed.
#     """
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Generating K8s resources for ws: {ws_name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(ws_name, refresh)
#     if pak8_conf is None:
#         pprint_error(f"Unable to parse config for {ws_name}")
#         return False
#
#     # Step 2: Get the K8s resource groups
#     k8s_resource_groups: Optional[
#         Dict[str, List[Dict[str, Any]]]
#     ] = get_k8s_resource_groups_as_dicts(
#         pak8_conf=pak8_conf, type_filters=type_filters, name_filters=name_filters
#     )
#     # This snippet receives the manifests from zues, currently commented out
#     # pak8_k8s_resources: Optional[Dict[str, List[Dict[str, Any]]]] = zeus_api.get_k8s_resource_groups_as_dicts(
#     #     id_workspace=ws_schema.id_workspace, pak8_conf=pak8_conf
#     # )
#
#     _resource_list: List[str] = []
#     _resources_str: str = ""
#     _resources_dir: Optional[Path] = config.get_k8s_resources_dir_path_by_ws_name(
#         ws_name
#     )
#     if k8s_resource_groups is None:
#         pprint_error("No K8s Resources available")
#         return False
#     elif _resources_dir is None:
#         pprint_error("Could not find path to store K8s resources")
#         return False
#
#     # delete directory to save resources if needed
#     if _resources_dir.exists():
#         if _resources_dir.is_file():
#             _resources_dir.unlink()
#         elif _resources_dir.is_dir():
#             delete_files_in_dir(_resources_dir)
#     _resources_dir.mkdir(exist_ok=True)
#
#     for rg_name, _rsrc_list in k8s_resource_groups.items():
#         # logger.debug("Processing K8sResourceGroup: {}\n{}".format(rg_name, _rsrc_list))
#         _rg_dir: Path = _resources_dir.joinpath(rg_name)
#         _rg_dir.mkdir(exist_ok=True)
#         for _rsrc_dict in _rsrc_list:
#             if _rsrc_dict:
#                 _fn_str = _rsrc_dict.get("metadata", {"name": "no_metadata_avl"}).get(
#                     "name", "no_name_avl"
#                 )
#                 _rsrc_file: Path = _rg_dir.joinpath(f"{_fn_str}.yaml")
#                 logger.debug(f"Creating {str(_rsrc_file)}")
#                 with open(_rsrc_file, "w") as _rsrc_f:
#                     yaml.dump(_rsrc_dict, _rsrc_f)
#
#     pprint_info(f"K8s resources saved under: {str(_resources_dir)}")
#     return True
#
#
# def print_k8s_resources_as_yaml(
#     ws_schema: schemas.WorkspaceSchema,
#     config: PhiConf,
#     refresh: bool = False,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> bool:
#     """Print the K8s resources for a workspace.
#     Similar to save but prints the resources instead
#     """
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Generating K8s resources for ws: {ws_name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(ws_name, refresh)
#     if pak8_conf is None:
#         pprint_error(f"Unable to parse config for {ws_name}")
#         return False
#
#     # Step 2: Get the K8s resources
#     pak8_k8s_resources: Optional[List[Dict[str, Any]]] = get_k8s_resources_as_dicts(
#         pak8_conf=pak8_conf, type_filters=type_filters, name_filters=name_filters
#     )
#     if pak8_k8s_resources is None:
#         pprint_error("No K8s resources available")
#         return False
#
#     print(yaml.dump_all(pak8_k8s_resources))
#     # pprint_info(pak8_k8s_resources)
#     # for _rsrc_dict in pak8_k8s_resources:
#     #     if _rsrc_dict:
#     #         pprint_info(yaml.dump(_rsrc_dict))
#     return True
#
#
# def apply_k8s_resources(
#     ws_schema: schemas.WorkspaceSchema,
#     config: PhiConf,
#     refresh: bool = False,
#     service_filters: Optional[List[str]] = None,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> bool:
#     """Apply the K8s resources for a workspace"""
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Applying K8s resources for ws: {ws_schema.name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name=ws_name, refresh=refresh, add_kubeconfig=True
#     )
#     if pak8_conf is None:
#         pprint_error(f"Unable to parse config for {ws_schema.name}")
#         return False
#
#     # Step 2: Create a GCPPak8 from the pak8_conf and validate
#     # TODO: Use Pak8Type
#     ws_pak8: Optional[GCPPak8] = None
#     try:
#         if pak8_conf.cloud_provider == pak8_enums.Pak8CloudProvider.GCP:
#             ws_pak8 = GCPPak8(pak8_conf)
#             logger.debug("GCPPak8 Created")
#             if ws_pak8 is None or not isinstance(ws_pak8, GCPPak8):
#                 logger.error("Invalid GCPPak8: {}".format(ws_pak8))
#                 return False
#         else:
#             logger.error(
#                 "Cloud Provider {} not yet supported".format(pak8_conf.cloud_provider)
#             )
#             return False
#     except pydantic.ValidationError as e:
#         logger.debug("e: {}".format(e))
#         error_msg = "Config Validation Errors. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return False
#
#     # Step 3: Get the current status of this Pak8
#     pak8_status: pak8_enums.Pak8Status = ws_pak8.get_pak8_status()
#     logger.debug("Pak8 is {}".format(pak8_status))
#
#     if pak8_status == pak8_enums.Pak8Status.CREATING_K8S_CLUSTER:
#         pprint_info("The cluster is being created, please try again in a few minutes")
#         return False
#     if not pak8_status.k8s_cluster_is_available():
#         pprint_info("No K8s Cluster available.")
#         return False
#
#     pak8_k8s_client: Optional[Pak8K8sClient] = ws_pak8.get_pak8_k8s_client()
#     if pak8_k8s_client is None:
#         pprint_error("K8s client unavailable.")
#         return False
#
#     # Step 4: After the GKEClusterSchema & K8sApiClient is available, we should have a PhiK8sData and kubeconfig
#     # If the PhiK8sData is not available, save the kubeconfig for future use
#     ws_k8s_data: Optional[PhiK8sData] = config.get_ws_k8s_data_by_name(ws_name)
#     if ws_k8s_data is None:
#         kconf_resource: Optional[Kubeconfig] = ws_pak8.get_kubeconfig()
#         if kconf_resource:
#             ws_k8s_data = config.update_ws_k8s_data(
#                 ws_name=ws_name, kubeconfig_resource=kconf_resource
#             )
#             if ws_k8s_data:
#                 pprint_status(f"Kubeconfig saved to: {ws_k8s_data.kubeconfig_path}")
#
#     # Step 5: Apply the K8s resources
#     return apply_k8s_resources_for_k8s_client(
#         pak8_conf=pak8_conf,
#         pak8_k8s_client=pak8_k8s_client,
#         service_filters=service_filters,
#         type_filters=type_filters,
#         name_filters=name_filters,
#     )
#
#
# def get_active_k8s_resources(
#     ws_schema: schemas.WorkspaceSchema,
#     config: PhiConf,
#     service_filters: Optional[List[str]] = None,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> Optional[Dict[str, List[Any]]]:
#     """Get active K8s resources for a workspace"""
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Getting active K8s resources for ws: {ws_schema.name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name=ws_name, refresh=False, add_kubeconfig=True
#     )
#     if pak8_conf is None:
#         pprint_error(f"Unable to parse config for {ws_schema.name}")
#         return None
#
#     # Step 2: Create a GCPPak8 from the pak8_conf and validate
#     # TODO: Use pak8.Pak8Type
#     ws_pak8: Optional[GCPPak8] = None
#     try:
#         if pak8_conf.cloud_provider == pak8_enums.Pak8CloudProvider.GCP:
#             ws_pak8 = GCPPak8(pak8_conf)
#             logger.debug("GCPPak8 Created")
#             if ws_pak8 is None or not isinstance(ws_pak8, GCPPak8):
#                 logger.error("Invalid GCPPak8: {}".format(ws_pak8))
#                 return None
#         else:
#             logger.error(
#                 "Cloud Provider {} not yet supported".format(pak8_conf.cloud_provider)
#             )
#             return None
#     except pydantic.ValidationError as e:
#         logger.debug("e: {}".format(e))
#         error_msg = "Config Validation Errors. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return None
#
#     # Step 3: Validate the current status of this Pak8
#     pak8_status: pak8_enums.Pak8Status = ws_pak8.get_pak8_status()
#     logger.debug("Pak8 is {}".format(pak8_status))
#
#     if pak8_status not in (
#         pak8_enums.Pak8Status.READY_TO_DEPLOY_K8S_RESOURCES,
#         pak8_enums.Pak8Status.K8S_RESOURCES_ACTIVE,
#     ):
#         logger.debug(f"returning because pak8_status: {pak8_status}")
#         return None
#
#     pak8_k8s_client: Optional[Pak8K8sClient] = ws_pak8.get_pak8_k8s_client()
#     if pak8_k8s_client is None:
#         pprint_error("K8s client unavailable.")
#         return None
#
#     # Step 4: After the GKEClusterSchema & K8sApiClient is available, we should have a PhiK8sData and kubeconfig
#     # If the PhiK8sData is not available, save the kubeconfig for future use
#     ws_k8s_data: Optional[PhiK8sData] = config.get_ws_k8s_data_by_name(ws_name)
#     if ws_k8s_data is None:
#         kconf_resource: Optional[Kubeconfig] = ws_pak8.get_kubeconfig()
#         if kconf_resource:
#             ws_k8s_data = config.update_ws_k8s_data(
#                 ws_name=ws_name, kubeconfig_resource=kconf_resource
#             )
#             if ws_k8s_data:
#                 pprint_status(f"Kubeconfig saved to: {ws_k8s_data.kubeconfig_path}")
#
#     # Step 5: Get the active K8s resources from the K8s Client
#     return get_active_k8s_resources_for_k8s_client(
#         pak8_conf=pak8_conf,
#         pak8_k8s_client=pak8_k8s_client,
#         service_filters=service_filters,
#         type_filters=type_filters,
#         name_filters=name_filters,
#     )
#
#
# def print_active_k8s_resources(
#     ws_schema: schemas.WorkspaceSchema,
#     config: PhiConf,
#     service_filters: Optional[List[str]] = None,
#     type_filters: Optional[List[str]] = None,
#     name_filters: Optional[List[str]] = None,
# ) -> None:
#     """Print active K8s objects for a workspace"""
#
#     active_k8s_resources: Optional[Dict[str, List[Any]]] = get_active_k8s_resources(
#         ws_schema=ws_schema,
#         config=config,
#         service_filters=service_filters,
#         type_filters=type_filters,
#         name_filters=name_filters,
#     )
#     if active_k8s_resources:
#         pprint_subheading("Active Resources: FIX ME")
#         # pprint_info(active_k8s_resources)
#     else:
#         pprint_subheading("No Active Resources")



from typing import Optional

from phidata.aws.config import AwsConfig
from phidata.aws.exceptions import AwsConfigException
from phidata.aws.manager import AwsManager

from phiterm.utils.cli_console import (
    print_error,
    print_heading,
    print_subheading,
    print_info,
)
from phidata.utils.log import logger


def deploy_aws_config(
    config: AwsConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the AwsManager
    aws_manager: AwsManager = config.get_aws_manager()
    if aws_manager is None:
        raise AwsConfigException("AwsManager unavailable")
    print_heading(f"--**-- Aws env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        aws_manager.create_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Create resources
    try:
        success: bool = aws_manager.create_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        raise

    # Step 4: Validate resources are created
    resource_creation_valid: bool = aws_manager.validate_resources_are_created(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resource_creation_valid:
        logger.error("AwsResource creation could not be validated")
        return False

    print_info("Aws config deployed")
    return True


def shutdown_aws_config(
    config: AwsConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the AwsManager
    aws_manager: AwsManager = config.get_aws_manager()
    if aws_manager is None:
        raise AwsConfigException("AwsManager unavailable")
    print_heading(f"--**-- Aws env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        aws_manager.delete_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Delete resources
    try:
        success: bool = aws_manager.delete_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        return False

    # Step 4: Validate resources are deleted
    resources_deletion_valid: bool = aws_manager.validate_resources_are_deleted(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resources_deletion_valid:
        logger.error("AwsResource deletion could not be validated")
        return False

    print_info("Aws config shutdown")
    return True


def patch_aws_config(
    config: AwsConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    app_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
) -> bool:
    # Step 1: Get the AwsManager
    aws_manager: AwsManager = config.get_aws_manager()
    if aws_manager is None:
        raise AwsConfigException("AwsManager unavailable")
    print_heading(f"--**-- Aws env: {config.env}\n")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        aws_manager.patch_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )
        return True

    # Step 3: Delete resources
    try:
        success: bool = aws_manager.patch_resources(
            name_filter=name_filter, type_filter=type_filter
        )
        if not success:
            return False
    except Exception as e:
        logger.error(e)
        raise

    # Step 4: Validate resources are patched
    resources_deletion_valid: bool = aws_manager.validate_resources_are_patched(
        name_filter=name_filter,
        type_filter=type_filter,
        app_filter=app_filter,
    )
    if not resources_deletion_valid:
        logger.error("AwsResource patch could not be validated")
        return False

    print_info("Aws config patched")
    return True



"""The Phi Cli

This is the entrypoint for the `phi` cli application.
"""

import typer

from phiterm.cli.ws.ws_app import ws_app
from phiterm.cli.k8s.k8s_app import k8s_app
from phiterm.cli.wf.wf_app import wf_app
from phiterm.cli.dx.dx_app import dx_app
from phiterm.utils.log import set_log_level_to_debug

cli_app = typer.Typer(
    help="""\b
Phidata is a toolkit for building data products as code.
\b
Get started:
1. Run `phi init` to init phidata on this machine
2. Run `phi ws init` to create a new workspace
3. Run `phi ws setup` to setup an existing workspace
4. Run `phi ws up` to deploy the active workspace
5. Run `phi ws down` to shutdown the active workspace
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="[command]",
    pretty_exceptions_show_locals=False,
)
cli_app.add_typer(ws_app)
cli_app.add_typer(k8s_app)
cli_app.add_typer(wf_app)
cli_app.add_typer(dx_app)


@cli_app.command(short_help="Initialize phidata, use -r to reset")
def init(
    reset: bool = typer.Option(
        False, "--reset", "-r", help="Reset phidata", show_default=True
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    login: bool = typer.Option(
        False, "--login", "-l", help="Login using phidata.com", show_default=True
    ),
):
    """
    \b
    Initialize phidata, use -r to reset

    \b
    Examples:
    * `phi init`    -> Initializing phidata
    * `phi init -r` -> Reset and initializing phidata
    """
    from phiterm.cli.cli_operator import initialize_phidata

    if print_debug_log:
        set_log_level_to_debug()

    init_success: bool = initialize_phidata(reset=reset, login=login)


@cli_app.command(short_help="Authenticate with phidata.com")
def auth(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Authenticate your account with phidata.
    """
    from phiterm.cli.cli_operator import authenticate_user

    if print_debug_log:
        set_log_level_to_debug()

    auth_success: bool = authenticate_user()


@cli_app.command(short_help="Set current directory as active workspace")
def set(
    ws_name: str = typer.Option(None, "-ws", help="Active workspace name"),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Setup the current directory as the active workspace.
    This command can be run from within the workspace directory
        OR with a -ws flag to set another workspace as primary.

    Set a workspace as active

    \b
    Examples:
    $ `phi ws set`           -> Set the current directory as the active phidata workspace
    $ `phi ws set -ws idata` -> Set the workspace named idata as the active phidata workspace
    """
    from phiterm.workspace.ws_operator import set_workspace_as_active

    if print_debug_log:
        set_log_level_to_debug()

    operation_success: bool = set_workspace_as_active(ws_name)


@cli_app.command(short_help="Log in from the cli")
def login(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Log in from the cli

    \b
    Examples:
    * `phi login`
    """
    from phiterm.cli.cli_operator import sign_in_using_cli

    if print_debug_log:
        set_log_level_to_debug()

    auth_success: bool = sign_in_using_cli()


@cli_app.command(short_help="Reset phidata installation", hidden=True)
def reset(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Reset the existing phidata installation
    After resetting please run `phi init` to initialize again.
    """
    from phiterm.cli.cli_operator import delete_phidata_conf

    if print_debug_log:
        set_log_level_to_debug()

    delete_phidata_conf()


@cli_app.command(short_help="Ping phidata servers", hidden=True)
def ping(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """Ping the phidata servers and check if you are authenticated"""
    from phiterm.api.user import user_ping, is_user_authenticated
    from phiterm.utils.cli_console import print_info

    if print_debug_log:
        set_log_level_to_debug()

    ping_success = user_ping()
    if ping_success:
        print_info("Ping successful")
    else:
        print_info("Could not reach phidata servers")

    if is_user_authenticated():
        print_info("User is authenticated")
    else:
        print_info("User is not authenticated, run `phi auth` to log in")


@cli_app.command(short_help="Print phidata config", hidden=True)
def config(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """Print your current phidata config"""
    from typing import Optional

    from phiterm.conf.phi_conf import PhiConf
    from phiterm.utils.cli_console import print_info

    if print_debug_log:
        set_log_level_to_debug()

    conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if conf is not None:
        conf.print_to_cli()
    else:
        print_info("Phidata has not been setup, run `phi init` to get started")


######################################################
## Notes
# * Documentation for invoke_without_command=True
#   https://typer.tiangolo.com/tutorial/commands/context/#executable-callback
# * To prevent re-wrapping during cli output, use \b before the paragraph.
#   More info: https://click.palletsprojects.com/en/5.x/documentation/#preventing-rewrapping
######################################################



from pathlib import Path
from typing import Optional, List, Dict

from phidata.workspace.config import WorkspaceConfig
from phiterm.workspace.ws_utils import get_ws_config_dir_path

from phiterm.utils.log import logger


def add_ws_dir_to_path(ws_root_path: Path) -> None:
    # Add ws_root_path to sys.path so is treated as a package
    try:
        import sys

        logger.debug(f"Adding {ws_root_path} to path")
        sys.path.insert(0, str(ws_root_path))
    except Exception as e:
        logger.warning(
            f"Could not add {ws_root_path} to Path. This will break workspace imports"
        )
        logger.exception(e)


def get_python_objects_from_module(module_path: Path) -> Dict:
    """
    Returns a list of python objects from a module
    """
    import importlib.util
    from importlib.machinery import ModuleSpec

    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    # Create a ModuleSpec
    module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location(
        "module", module_path
    )
    # Using the ModuleSpec create a module
    if module_spec:
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)  # type: ignore
        return module.__dict__
    else:
        return {}


def load_workspace(ws_root_path: Path) -> WorkspaceConfig:
    """
    Loads a workspace into a WorkspaceConfig
    """

    ws_config: WorkspaceConfig = WorkspaceConfig()
    config_dir_path: Optional[Path] = get_ws_config_dir_path(ws_root_path)
    if config_dir_path is not None:
        logger.debug(f"--^^-- Loading workspace from: {config_dir_path}")
        workspace_config_objects = {}
        config_files = config_dir_path.rglob("*.py")
        for _config_file in config_files:
            if _config_file.name == "__init__.py":
                continue
            logger.debug(f"Reading file: {_config_file}")
            try:
                python_objects = get_python_objects_from_module(_config_file)
                for obj_name, obj in python_objects.items():
                    _type_name = obj.__class__.__name__
                    if _type_name in [
                        "WorkspaceSettings",
                        "DockerConfig",
                        "K8sConfig",
                        "AwsConfig",
                    ]:
                        workspace_config_objects[obj_name] = obj
            except Exception as e:
                pass

        logger.debug(f"workspace_config_objects: {workspace_config_objects}")
        for obj_name, obj in workspace_config_objects.items():
            _obj_type = obj.__class__.__name__
            logger.debug(f"Adding {obj_name}: | Type: {_obj_type}")
            if _obj_type == "WorkspaceSettings":
                ws_config.ws_settings = obj
                try:
                    ws_config.default_env = obj.dev_env
                except Exception as e:
                    pass
                try:
                    ws_config.aws_region = obj.aws_region
                except Exception as e:
                    pass
            elif _obj_type == "DockerConfig":
                if ws_config.docker is None:
                    ws_config.docker = []
                ws_config.docker.append(obj)
                ws_config.default_config = "docker"
            elif _obj_type == "K8sConfig":
                if ws_config.k8s is None:
                    ws_config.k8s = []
                ws_config.k8s.append(obj)
            elif _obj_type == "AwsConfig":
                if ws_config.aws is None:
                    ws_config.aws = []
                ws_config.aws.append(obj)

    return ws_config



from typing import Optional

from typer import launch as typer_launch

from phiterm.api.exceptions import CliAuthException
from phiterm.api.user import (
    sign_in_user,
    authenticate_and_get_user,
)
from phiterm.conf.constants import PHI_CONF_DIR, PHI_SIGNIN_URL_WITHOUT_PARAMS
from phiterm.conf.phi_conf import PhiConf
from phiterm.schemas.user import UserSchema, EmailPasswordSignInSchema
from phiterm.utils.cli_auth_server import (
    get_port_for_auth_server,
    get_auth_token_from_web_flow,
)
from phiterm.utils.cli_console import (
    print_error,
    print_heading,
    print_info,
)
from phiterm.utils.filesystem import delete_from_fs
from phiterm.utils.log import logger


def delete_phidata_conf() -> None:
    logger.debug("Removing existing Phidata configuration")
    delete_from_fs(PHI_CONF_DIR)


def authenticate_user() -> bool:
    """Authenticate the user using credentials from phidata.com
    Steps:
    1. Authenticate the user by opening the phidata sign-in url
        and the web-app will post an auth token to a mini http server
        running on the auth_server_port.
    2. Using the auth_token, authenticate the CLI with backend and
        save the auth_token to PHI_AUTH_TOKEN_PATH.
        This step is handled by authenticate_and_get_user()
    5. After the user is authenticated create a PhiConf if needed.
    """

    print_heading("Logging in at phidata.com ...")

    auth_server_port = get_port_for_auth_server()
    redirect_uri = "http%3A%2F%2Flocalhost%3A{}%2F".format(auth_server_port)
    auth_url = "{}?source=cli&action=signin&redirecturi={}".format(
        PHI_SIGNIN_URL_WITHOUT_PARAMS, redirect_uri
    )
    print_info("\nYour browser will be opened to visit:\n{}".format(auth_url))
    typer_launch(auth_url)
    print_info("\nWaiting for a response from browser...\n")

    tmp_auth_token = get_auth_token_from_web_flow(auth_server_port)
    if tmp_auth_token is None:
        print_error(f"Could not authenticate, please run `phi auth` again")
        return False

    try:
        user: Optional[UserSchema] = authenticate_and_get_user(tmp_auth_token)
    except CliAuthException as e:
        logger.exception(e)
        print_error(f"Could not authenticate, please run `phi auth` again")
        return False

    if user is None:
        print_error(f"Could not get user data, please run `phi auth` again")
        return False

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        phi_conf = PhiConf(user)

    phi_conf.user = user
    print_info("Welcome {}, you are now logged in\n".format(user.email))

    return phi_conf.sync_workspaces_from_api()


def initialize_phidata(reset: bool = False, login: bool = False) -> bool:
    """This function is called by `phi init` and initializes phidata on the users machine.
    Steps:
    1. Check if PHI_CONF_DIR exists, if not, create it. If reset == True, recreate PHI_CONF_DIR.
    2. Check if PhiConf exists, if it does, try and authenticate the user
        If the user is authenticated, phi is configured and authenticated. Return True.
    3. If PhiConf does not exist, create a new PhiConf. Return True.
    """

    print_heading("Welcome to phidata!\n")
    if reset:
        delete_phidata_conf()

    logger.debug("Initializing phidata")

    # Check if ~/.phi exists, if it is not a dir - delete it and create the dir
    if PHI_CONF_DIR.exists():
        logger.debug(f"{PHI_CONF_DIR} exists")
        if not PHI_CONF_DIR.is_dir():
            try:
                delete_from_fs(PHI_CONF_DIR)
            except Exception as e:
                logger.exception(e)
                print_info(
                    f"Something went wrong, please delete {PHI_CONF_DIR} and run `phi init` again"
                )
                return False
            PHI_CONF_DIR.mkdir(parents=True)
    else:
        PHI_CONF_DIR.mkdir(parents=True)
        logger.debug(f"created {PHI_CONF_DIR}")

    # Confirm PHI_CONF_DIR exists otherwise we should return
    if PHI_CONF_DIR.exists():
        logger.debug(f"Your phidata config is stored at: {PHI_CONF_DIR}")
    else:
        print_info(f"Something went wrong, please run `phi init` again")
        return False

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        # Create a new PhiConf
        phi_conf = PhiConf()

    # Authenticate user
    if login:
        auth_success = authenticate_user()

    if phi_conf is not None:
        print_info("Phidata is initialized. Next steps:\n")
        print_info(" 1. Run `phi ws init` to create a new workspace")
        print_info(" 2. Run `phi ws setup` to setup an existing workspace")
        print_info(" 3. Run `phi ws up -dr` to dry-run workspace deploy")
        print_info(" 4. Run `phi ws up` to deploy the workspace")
    return True


def sign_in_using_cli() -> bool:
    from getpass import getpass

    print_heading("Log in")
    email_raw = input("email: ")
    pass_raw = getpass()

    if email_raw is None or pass_raw is None:
        print_error("Incorrect email or password")

    try:
        user: Optional[UserSchema] = sign_in_user(
            EmailPasswordSignInSchema(email=email_raw, password=pass_raw)
        )
    except CliAuthException as e:
        logger.exception(e)
        return False

    if user is None:
        print_error("Could not get user data, please log in again")
        return False

    print_info("Welcome {}, you are now authenticated\n".format(user.email))

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if phi_conf is None:
        phi_conf = PhiConf()

    phi_conf.user = user
    return phi_conf.sync_workspaces_from_api()



import json
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional

from phiterm.conf.constants import PHI_AUTH_TOKEN_PATH
from phiterm.utils.log import logger


class CliAuthRequestHandler(BaseHTTPRequestHandler):
    """Request Handler to accept the CLI auth token after the web based auth flow.
    Reference: https://medium.com/@hasinthaindrajee/browser-sso-for-cli-applications-b0be743fa656
    Source: https://gist.github.com/mdonkers/63e115cc0c79b4f6b8b3a6b797e485c7
    TODO:
        * Fix the header and limit to only localhost or phidata.com
    """

    def _set_response(self):
        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Headers", "*")
        self.send_header("Access-Control-Allow-Methods", "POST")
        self.end_headers()

    # def do_GET(self):
    #     logger.info("GET request,\nPath: %s\nHeaders:\n%s\n", str(self.path), str(self.headers))
    #     self._set_response()
    #     self.wfile.write("GET request for {}".format(self.path).encode('utf-8'))

    def do_OPTIONS(self):
        # logger.info(
        #     "OPTIONS request,\nPath: %s\nHeaders:\n%s\n",
        #     str(self.path),
        #     str(self.headers),
        # )
        self._set_response()
        # self.wfile.write("OPTIONS request for {}".format(self.path).encode('utf-8'))

    def do_POST(self):
        content_length = int(
            self.headers["Content-Length"]
        )  # <--- Gets the size of data
        post_data = self.rfile.read(content_length)  # <--- Gets the data itself
        # logger.info(
        #     "POST request,\nPath: {}\nHeaders:\n{}\n\nBody:\n{}\n".format(
        #         str(self.path), str(self.headers), post_data.decode("utf-8")
        #     )
        # )
        # logger.info("Data: {}".format(post_data))
        # logger.info("type: {}".format(type(post_data)))
        PHI_AUTH_TOKEN_PATH.touch(exist_ok=True)
        PHI_AUTH_TOKEN_PATH.write_text(post_data.decode("utf-8"))
        # TODO: Add checks before shutting down the server
        self.server.running = False  # type: ignore
        self._set_response()

    def log_message(self, format, *args):
        pass


class CliAuthServer:
    """
    Source: https://stackoverflow.com/a/38196725/10953921
    """

    def __init__(self, port=9190):
        self._server = HTTPServer(("", port), CliAuthRequestHandler)
        self._thread = threading.Thread(target=self.run)
        self._thread.daemon = True
        self._server.running = False  # type: ignore

    def run(self):
        self._server.running = True  # type: ignore
        while self._server.running:  # type: ignore
            self._server.handle_request()

    def start(self):
        self._thread.start()

    def shut_down(self):
        self._thread.close()  # type: ignore


def get_port_for_auth_server():
    # TODO: Check if port is available
    return 9190


def get_auth_token_from_web_flow(port) -> Optional[str]:
    """
    GET request: curl http://localhost:9190
    POST request: curl -d "foo=bar&bin=baz" http://localhost:9190
    """

    server = CliAuthServer(port)
    server.run()

    if PHI_AUTH_TOKEN_PATH.exists() and PHI_AUTH_TOKEN_PATH.is_file():
        auth_token_str = PHI_AUTH_TOKEN_PATH.read_text()
        auth_token_json = json.loads(auth_token_str)
        PHI_AUTH_TOKEN_PATH.unlink()
        return auth_token_json.get("AuthToken", None)
    return None
